
local addonName, addon = ...
local KT = LibStub("AceAddon-3.0"):GetAddon(addonName)
KT.title = GetAddOnMetadata(addonName, "Title")
KT.version = GetAddOnMetadata(addonName, "Version")
local QGC_LOADMSG = "Activated |cFFFFA500"KT.title..."v"KT.version.."|r - |cFFFFD700Lazzy-Kilrogg(US)|r"
local LSM = LibStub("LibSharedMedia-3.0")
local MSQ = LibStub("Masque", true)
local _DBG = function(...) if _DBG then _DBG("KT", ...) end end

local _G = _G

local floor = math.floor
local format = string.format
local hooksecurefunc = hooksecurefunc
local pairs = pairs
local strfind = string.find
local tinsert = table.insert
local tremove = table.remove
local unpack = unpack
local round = function(n) return floor(n + 0.5) end

local CreateFrame = CreateFrame
local InCombatLockdown = InCombatLockdown
local UIParent = UIParent
local GameTooltip = GameTooltip

local trackerWidth = 280
local paddingBottom = 16
	
local fixedButtons = {}
local freeTags = {}
local freeButtons = {}
	
local stopUpdate = false
local activeTask = 0
local animTask = false

local db, hbColor, font, fontShadow 
local IsPetTracker

-- Main frame
local KTF = CreateFrame("Frame", addonName.."Frame", UIParent)
KT.frame = KTF

-- Blizzard OTF objects
local OTF = ObjectiveTrackerFrame
local OTFHeader = OTF.HeaderMenu
local OTFButton = OTF.HeaderMenu.MinimizeButton

OBJECTIVE_TRACKER_COLOR = {
	["Normal"] = { r = 0.8, g = 0.8, b = 0.8 },
	["NormalHighlight"] = { r = HIGHLIGHT_FONT_COLOR.r, g = HIGHLIGHT_FONT_COLOR.g, b = HIGHLIGHT_FONT_COLOR.b },
	["Failed"] = { r = DIM_RED_FONT_COLOR.r, g = DIM_RED_FONT_COLOR.g, b = DIM_RED_FONT_COLOR.b },
	["FailedHighlight"] = { r = RED_FONT_COLOR.r, g = RED_FONT_COLOR.g, b = RED_FONT_COLOR.b },
	["Header"] = { r = 1, g = 0.5, b = 0 },		-- orange
	["HeaderHighlight"] = { r = NORMAL_FONT_COLOR.r, g = NORMAL_FONT_COLOR.g, b = NORMAL_FONT_COLOR.b },
	["Complete"] = { r = 0, g = 1, b = 0 },		-- green
}

--------------------
-- Load functions --
--------------------

function KT:OnInitialize()
	-- Get character data
	local _, class = UnitClass("player")
	self.classColor = RAID_CLASS_COLORS[class]
	
	-- Setup Options
	self:SetupOptions()
	db = self.db.profile
end

function KT:OnEnable()
	-- Supported Addons
	self:PetTracker_SetSupport()
	self:Masque_SetSupport()
	self:ElvUI_SetSupport()
	self:Tukui_SetSupport()
	
	self.screenWidth = floor(GetScreenWidth())
	self.screenHeight = floor(GetScreenHeight())
	
	OTFHeader:Show()
	OTFHeader.Hide = function() end
	OTFHeader.SetShown = function() end
	OTFHeader.Title:ClearAllPoints()
	OTFHeader.Title:SetPoint("LEFT", OTF, "LEFT", -20, 0)
	
	KTF:RegisterEvent("PLAYER_ENTERING_WORLD")
	KTF:SetScript("OnEvent", function()
		self:SetHooks()
		self:PetTracket_SetHooks()
		
		self:SetFrames()
		self:Init()
		print(QGC_LOADMSG)
		if db.collapsed then
			ObjectiveTracker_MinimizeButton_OnClick()
		end
	end)
end

--------------------
-- Init functions --
--------------------

function KT:Init()
	db = self.db.profile
	
	if db.keyBindMinimize ~= "" then
		SetOverrideBindingClick(KTF, false, db.keyBindMinimize, OTFButton:GetName())
	end
	
	self:MoveTracker()
	self:SetBackground()
	self:SetText()
	ObjectiveTracker_Update()
end

function KT:SetFrames()
	-- Main frame
	KTF:SetWidth(trackerWidth)
	KTF:SetFrameStrata(db.frameStrata)
	KTF:SetFrameLevel(KTF:GetFrameLevel() + 25)
	
	if not db.collapseInInstance then
		KTF:UnregisterEvent("PLAYER_ENTERING_WORLD")
	end
	KTF:RegisterEvent("QUEST_WATCH_LIST_CHANGED")
	KTF:RegisterEvent("TRACKED_ACHIEVEMENT_LIST_CHANGED")
	KTF:RegisterEvent("QUEST_ACCEPTED")
	KTF:RegisterEvent("QUEST_AUTOCOMPLETE")
	KTF:RegisterEvent("PLAYER_REGEN_ENABLED")
	KTF:RegisterEvent("ZONE_CHANGED")
	KTF:RegisterEvent("ZONE_CHANGED_NEW_AREA")
	KTF:RegisterEvent("QUEST_POI_UPDATE")
	KTF:RegisterEvent("WORLD_MAP_UPDATE")
	KTF:RegisterEvent("UPDATE_BINDINGS")
	--[[ KTF:RegisterEvent("QUEST_LOG_UPDATE")
	KTF:RegisterEvent("SCENARIO_UPDATE")
	KTF:RegisterEvent("SCENARIO_CRITERIA_UPDATE")
	KTF:RegisterEvent("TRACKED_ACHIEVEMENT_UPDATE")
	KTF:RegisterEvent("VARIABLES_LOADED")
	KTF:RegisterEvent("QUEST_TURNED_IN")
	KTF:RegisterEvent("PET_JOURNAL_LIST_UPDATE") ]]

	KTF:SetScript("OnEvent", function(_, event, ...)
		_DBG("Event - "..event)
		if event == "PLAYER_ENTERING_WORLD" then
			if IsInInstance() and not db.collapsed then
				ObjectiveTracker_MinimizeButton_OnClick()
			end
		elseif event == "QUEST_WATCH_LIST_CHANGED" or event == "TRACKED_ACHIEVEMENT_LIST_CHANGED" then
			local questID, added = ...
			KT:ToggleEmptyTracker(added)
		elseif event == "QUEST_ACCEPTED" then
			local questLogIndex, questID = ...
		elseif event == "QUEST_AUTOCOMPLETE" then
			if db.collapsed then
				ObjectiveTracker_MinimizeButton_OnClick()
			end
		elseif event == "PLAYER_REGEN_ENABLED" and stopUpdate then
			stopUpdate = false
			RemoveFixedButton()
			if not db.collapsed then
				ObjectiveTracker_Update(OBJECTIVE_TRACKER_UPDATE_QUEST)
			end
		elseif (event == "ZONE_CHANGED" or event == "ZONE_CHANGED_NEW_AREA") and KTF.Buttons.num > 0 then
			KTF.Buttons.reanchor = not InCombatLockdown()
		elseif event == "UPDATE_BINDINGS" then
			QuestActiveButton_UpdateHotkey()
		end
		if event == "QUEST_WATCH_LIST_CHANGED" or event == "QUEST_POI_UPDATE" or event == "ZONE_CHANGED" or event == "WORLD_MAP_UPDATE" then
			self:UpdateActiveButton()
		end
	end)
	
	local button
	-- Minimize button
	OTFButton:Hide()
	button = CreateFrame("Button", addonName.."MinimizeButton", KTF)
	button:SetSize(16, 16)
	button:SetPoint("TOPRIGHT", -10, -8)
	button:SetFrameLevel(KTF:GetFrameLevel() + 10)
	button:SetNormalTexture("Interface\\AddOns\\"..addonName.."\\Media\\btns_header")
	button:GetNormalTexture():SetTexCoord(0, 0.5, 0.25, 0.5)
		
	button:RegisterForClicks("AnyDown")
	button:SetScript("OnClick", function(self, btn)
		if IsAltKeyDown() then
			KT:OpenOptions()
		elseif IsControlKeyDown() then
			KT:ToggleTrackAllQuests()
		elseif not KT:IsTrackerEmpty() then
			ObjectiveTracker_MinimizeButton_OnClick()
		end
	end)
	button:SetScript("OnEnter", function(self)
		self:GetNormalTexture():SetVertexColor(1, 1, 1)
		GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
		local title = db.keyBindMinimize ~= "" and KT.title.." "..NORMAL_FONT_COLOR_CODE.."("..db.keyBindMinimize..")|r" or KT.title
		GameTooltip:AddLine(title, 1, 1, 1)
		GameTooltip:AddLine("Ctrl+Click - track/untrack all quests", 0.5, 0.5, 0.5)
		GameTooltip:AddLine("Alt+Click - addon Options", 0.5, 0.5, 0.5)
		GameTooltip:Show()
	end)
	button:SetScript("OnLeave", function(self)
		self:GetNormalTexture():SetVertexColor(hbColor.r, hbColor.g, hbColor.b)
		GameTooltip:Hide()
	end)
	OTFButton = button
	OTF.HeaderMenu.MinimizeButton = button
	
	if db.hbOtherButtons then
		-- Quest Log button
		button = CreateFrame("Button", addonName.."QuestLogButton", KTF)
		button:SetSize(16, 16)
		button:SetPoint("TOPRIGHT", -50, -8)
		button:SetFrameLevel(KTF:GetFrameLevel() + 10)
		button:SetNormalTexture("Interface\\AddOns\\"..addonName.."\\Media\\btns_header")
		button:GetNormalTexture():SetTexCoord(0.5, 1, 0, 0.25)
		
		button:RegisterForClicks("AnyDown")
		button:SetScript("OnClick", function(self, btn)
			ToggleQuestLog()
		end)
		button:SetScript("OnEnter", function(self)
			self:GetNormalTexture():SetVertexColor(1, 1, 1)
			GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
			GameTooltip:AddLine(QuestLogMicroButton.tooltipText, 1, 1, 1)
			GameTooltip:Show()
		end)
		button:SetScript("OnLeave", function(self)
			self:GetNormalTexture():SetVertexColor(hbColor.r, hbColor.g, hbColor.b)
			GameTooltip:Hide()
		end)
		OTFHeader.QuestLogButton = button
		
		-- Achievements button
		button = CreateFrame("Button", addonName.."AchievementsButton", KTF)
		button:SetSize(16, 16)
		button:SetPoint("TOPRIGHT", -30, -8)
		button:SetFrameLevel(KTF:GetFrameLevel() + 10)
		button:SetNormalTexture("Interface\\AddOns\\"..addonName.."\\Media\\btns_header")
		button:GetNormalTexture():SetTexCoord(0.5, 1, 0.25, 0.5)
		
		button:RegisterForClicks("AnyDown")
		button:SetScript("OnClick", function(self, btn)
			ToggleAchievementFrame()
		end)
		button:SetScript("OnEnter", function(self)
			self:GetNormalTexture():SetVertexColor(1, 1, 1)
			GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
			GameTooltip:AddLine(AchievementMicroButton.tooltipText, 1, 1, 1)
			GameTooltip:Show()
		end)
		button:SetScript("OnLeave", function(self)
			self:GetNormalTexture():SetVertexColor(hbColor.r, hbColor.g, hbColor.b)
			GameTooltip:Hide()
		end)
		OTFHeader.AchievementsButton = button
	end
	
	-- Scroll frame
	local Scroll = CreateFrame("ScrollFrame", addonName.."Scroll", KTF)
	Scroll:SetPoint("TOPLEFT", 34, -5)
	Scroll:SetPoint("BOTTOMRIGHT", -4, 5)
	--Scroll:SetHitRectInsets(-30, 0, 0, 1)
	Scroll:EnableMouseWheel(true)
	Scroll.step = 20
	Scroll.value = 0
	Scroll:SetScript("OnMouseWheel", function(self, delta)
		if not db.collapsed and OTF.height > db.maxHeight then	-- floats ??? +11
			if delta < 0 then
				self.value = (self.value+self.step < OTF.height-db.maxHeight) and self.value + self.step or OTF.height - db.maxHeight
			else
				self.value = (self.value-self.step > 0) and self.value - self.step or 0
			end
			self:SetVerticalScroll(self.value)
			if self.value > 0 and self.value < OTF.height-db.maxHeight then
				CloseDropDownMenus()
			end				
			_DBG("SCROLL ... "..self.value.." ... "..OTF.height.." - "..db.maxHeight)
		end
	end)
	Scroll:SetScript("OnScrollRangeChanged", function(self, x, y)
		KT:MoveButtons()
	end)
	KTF.Scroll = Scroll

	-- Blizzard frame
	OTF:SetParent(KTF.Scroll)
	OTF:SetClampedToScreen(false)
	OTF:SetMovable(false)
	OTF.SetMovable = function() end
	OTF:ClearAllPoints()
	OTF.ClearAllPoints = function() end
	OTF:SetPoint("TOPLEFT")
	OTF:SetPoint("BOTTOMRIGHT")
	OTF.SetPoint = function() end
	KTF.Scroll:SetScrollChild(OTF)
	
	-- Buttons frame
	local Buttons = CreateFrame("Frame", addonName.."Buttons", UIParent)
	Buttons:SetSize(40, 40)
	Buttons:SetPoint("TOPLEFT", 0, 0)
	Buttons:SetFrameStrata(db.frameStrata)
	Buttons:SetFrameLevel(KTF:GetFrameLevel() - 1)
	Buttons:SetAlpha(0)
	
	Buttons.num = 0
	Buttons.height = 0
	Buttons.timer = 0
	Buttons.timerID = nil
	Buttons.reanchor = false
	KTF.Buttons = Buttons
	
	-- Active (closest) button frame
	local name = addonName.."ActiveButton"
	local abutton = CreateFrame("Button", name, ExtraActionBarFrame, "SecureActionButtonTemplate")
	abutton:SetSize(52, 52)
	abutton:SetPoint("CENTER", 0, 0.5)
	
	abutton.icon = abutton:CreateTexture(name.."Icon", "BACKGROUND")
	abutton.icon:SetPoint("TOPLEFT", 0, -1)
	abutton.icon:SetPoint("BOTTOMRIGHT", 0, -1)
	
	abutton.Style = abutton:CreateTexture(name.."Style", "OVERLAY")
	abutton.Style:SetSize(256, 128)
	abutton.Style:SetPoint("CENTER", -2, 0)
	abutton.Style:SetTexture("Interface\\ExtraButton\\Default")
	
	abutton.Count = abutton:CreateFontString(name.."Count", "OVERLAY", "NumberFontNormal")
	abutton.Count:SetJustifyH("RIGHT")
	abutton.Count:SetPoint("BOTTOMRIGHT", abutton.icon, -2, 2)
	
	abutton.Cooldown = CreateFrame("Cooldown", name.."Cooldown", abutton, "CooldownFrameTemplate")
	abutton.Cooldown:ClearAllPoints()
	abutton.Cooldown:SetPoint("TOPLEFT", 4, -4)
	abutton.Cooldown:SetPoint("BOTTOMRIGHT", -3, 2)
	
	abutton.HotKey = abutton:CreateFontString(name.."HotKey", "ARTWORK", "NumberFontNormalSmallGray")
	abutton.HotKey:SetSize(30, 10)
	abutton.HotKey:SetJustifyH("RIGHT")
	abutton.HotKey:SetText(RANGE_INDICATOR)
	abutton.HotKey:SetPoint("TOPRIGHT", abutton.icon, -2, -7)
	
	abutton.text = abutton:CreateFontString(name.."Text", "ARTWORK", "NumberFontNormalSmall")
	abutton.text:SetSize(20, 10)
	abutton.text:SetJustifyH("LEFT")
	abutton.text:SetPoint("TOPLEFT", abutton.icon, 4, -7)
	
	abutton:SetScript("OnEvent", QuestObjectiveItem_OnEvent)
	abutton:SetScript("OnUpdate", QuestObjectiveItem_OnUpdate)
	abutton:SetScript("OnShow", QuestObjectiveItem_OnShow)
	abutton:SetScript("OnHide", QuestObjectiveItem_OnHide)
	abutton:SetScript("OnEnter", QuestObjectiveItem_OnEnter)
	abutton:SetScript("OnLeave", QuestObjectiveItem_OnLeave)
	abutton:RegisterForClicks("AnyUp")
	abutton:SetAttribute("type","item")
	
	abutton:SetPushedTexture("Interface\\Buttons\\UI-Quickslot-Depress")
	do local tex = abutton:GetPushedTexture()
		tex:SetPoint("TOPLEFT", 0, -1)
		tex:SetPoint("BOTTOMRIGHT", 0, -1)
	end
	abutton:SetHighlightTexture("Interface\\Buttons\\ButtonHilight-Square", "ADD")
	do local tex = abutton:GetHighlightTexture()
		tex:SetPoint("TOPLEFT", 0, -1)
		tex:SetPoint("BOTTOMRIGHT", 0, -1)
	end
	abutton:Hide()
	KTF.ActiveButton = abutton
	self:Masque_AddButton(abutton)
	
	-- Frame locks
	if FRAMELOCK_STATES and FRAMELOCK_STATES.PETBATTLES then
		FRAMELOCK_STATES.PETBATTLES.KalielsTrackerFrame = "hidden"
		FRAMELOCK_STATES.PETBATTLES.KalielsTrackerButtons = "hidden"
	end
end

function KT:MoveTracker()
	KTF:ClearAllPoints()
	KTF:SetPoint(db.anchorPoint, UIParent, db.anchorPoint, db.xOffset, db.yOffset)
	KTF.directionUp = (db.anchorPoint == "BOTTOMLEFT" or db.anchorPoint == "BOTTOMRIGHT")
	KTF.anchorLeft = (db.anchorPoint == "LEFT" or db.anchorPoint == "TOPLEFT" or db.anchorPoint == "BOTTOMLEFT")

	local options = self.options.args.general.args.sec1.args
	if KTF.anchorLeft then
		options.xOffset.min = 0
		options.xOffset.max = self.screenWidth - trackerWidth
	else
		options.xOffset.min = -(self.screenWidth - trackerWidth)
		options.xOffset.max = 0
	end

	if KTF.directionUp then
		options.yOffset.min = 0
		options.yOffset.max = self.screenHeight - 100
	else
		options.yOffset.min = -(self.screenHeight - 100)
		options.yOffset.max = 0
	end

	local mod = (db.anchorPoint == "LEFT" or db.anchorPoint == "RIGHT") and abs(db.yOffset) or 0
	options.maxHeight.max = self.screenHeight - abs(db.yOffset) - mod
	db.maxHeight = (abs(db.yOffset)+db.maxHeight > self.screenHeight-mod) and options.maxHeight.max or db.maxHeight

	self:MoveButtons()
end

function KT:SetBackground()
	local backdrop = {
		bgFile = LSM:Fetch("background", db.bgr),
		edgeFile = LSM:Fetch("border", db.border),
		edgeSize = db.borderThickness,
		insets = { left=db.bgrInset, right=db.bgrInset, top=db.bgrInset, bottom=db.bgrInset }
	}
	local borderColor = db.classBorder and self.classColor or db.borderColor
	
	KTF:SetBackdrop(backdrop)
	KTF:SetBackdropColor(db.bgrColor.r, db.bgrColor.g, db.bgrColor.b, db.bgrColor.a)
	KTF:SetBackdropBorderColor(borderColor.r, borderColor.g, borderColor.b, db.borderAlpha)
	
	hbColor = db.hbBorderColor and borderColor or db.hbColor
	OTFButton:GetNormalTexture():SetVertexColor(hbColor.r, hbColor.g, hbColor.b)
	if db.hbOtherButtons then
		OTFHeader.QuestLogButton:GetNormalTexture():SetVertexColor(hbColor.r, hbColor.g, hbColor.b)
		OTFHeader.AchievementsButton:GetNormalTexture():SetVertexColor(hbColor.r, hbColor.g, hbColor.b)
	end
	
	if db.qiBgrBorder then
		KTF.Buttons:SetBackdrop(backdrop)
		KTF.Buttons:SetBackdropColor(db.bgrColor.r, db.bgrColor.g, db.bgrColor.b, db.bgrColor.a)
		KTF.Buttons:SetBackdropBorderColor(borderColor.r, borderColor.g, borderColor.b, db.borderAlpha)
	else
		KTF.Buttons:SetBackdrop(nil)
	end
	
	if db.bgr == "None" and db.border == "None" then
		KTF.Scroll:SetHitRectInsets(-150, 0, 0, 1)
	else
		KTF.Scroll:SetHitRectInsets(-30, 0, 0, 1)
	end
end

function KT:SetText()
	font = LSM:Fetch("font", db.font)
	fontShadow = db.fontShadow and 1 or 0
	
	-- Title (collapsed)
	OTFHeader.Title:SetFont(font, db.fontSize, db.fontFlag)
	OTFHeader.Title:SetTextColor(0, 0.7, 1)	-- blue
	OTFHeader.Title:SetShadowColor(0, 0, 0, fontShadow)
	--Quests
	self:SetHeader(ObjectiveTrackerBlocksFrame.QuestHeader)		
	-- Achievements
	self:SetHeader(ObjectiveTrackerBlocksFrame.AchievementHeader)
	-- Scenario
	self:SetHeader(ObjectiveTrackerBlocksFrame.ScenarioHeader)
	ScenarioStageBlock.Stage:SetFont(font, db.fontSize+5, db.fontFlag)
	-- Bonus objective
	self:SetHeader(BONUS_OBJECTIVE_TRACKER_MODULE.Header)
	-- Addon - PetTracker
	if IsPetTracker and PetTracker.Objectives.Header then
		self:SetHeader(PetTracker.Objectives.Header)
	end
end

function KT:SetSize()
	local height = 33	-- collapsed
	local mod = 0
	local PTheight = 0
	
	if OTF.BlocksFrame.contentsHeight then
		OTF.BlocksFrame.contentsHeight = round(OTF.BlocksFrame.contentsHeight)
	else
		return
	end
	
	_DBG(" - height = "..OTF.BlocksFrame.contentsHeight)
	if not db.collapsed and not self:IsTrackerEmpty() then
		if AUTO_QUEST_POPUP_TRACKER_MODULE.firstBlock then
			mod = mod + (GetNumAutoQuestPopUps() * 68)
		end
		if BONUS_OBJECTIVE_TRACKER_MODULE.firstBlock then
			mod = mod + BONUS_OBJECTIVE_TRACKER_MODULE.blockPadding
		end
		if self:PetTracker_IsShown() then
			PTheight = round(PetTracker.Objectives:GetHeight()) + 6 - (self:IsTrackerEmpty(true) and paddingBottom or 0)
			mod = mod + PTheight
		end
		height = OBJECTIVE_TRACKER_HEADER_HEIGHT + OTF.BlocksFrame.contentsHeight + mod
		_DBG(" - "..OBJECTIVE_TRACKER_HEADER_HEIGHT.." + "..OTF.BlocksFrame.contentsHeight.." + "..mod.." = "..height)
		OTF:SetHeight(height)
		OTF.height = height
		
		if height < 2000 and KTF.Scroll:GetHeight() < 2000 then
			if height > db.maxHeight then
				_DBG("MOVE ... "..KTF.Scroll.value.." > "..OTF.height.." - "..db.maxHeight)
				if KTF.Scroll.value > OTF.height-db.maxHeight then
					KTF.Scroll.value = OTF.height - db.maxHeight
				end
				KTF.Scroll:SetVerticalScroll(KTF.Scroll.value)
				height = db.maxHeight
			elseif height <= db.maxHeight then
				if KTF.Scroll.value > 0 then
					KTF.Scroll.value = 0
					KTF.Scroll:SetVerticalScroll(0)
				end
				KTF.Scroll:UpdateScrollChildRect()
			end
			self:MoveButtons()
		end
	else
		OTF.height = height - 10		
		OTF:SetHeight(OTF.height)
		KTF.Scroll:SetVerticalScroll(0)
	end
	
	KTF:SetHeight(height)
	KTF.height = height
end

function KT:MoveButtons()
	if not InCombatLockdown() then
		local point, xOfs, yOfs
		if KTF.anchorLeft then
			point = "LEFT"
			xOfs = KTF:GetRight() and KTF:GetRight() + db.qiXOffset
		else
			point = "RIGHT"
			xOfs = KTF:GetLeft() and KTF:GetLeft() - db.qiXOffset
		end
		local yMod = 2 + (4 - db.bgrInset)
		if KTF.directionUp and db.maxHeight < KTF.Buttons:GetHeight() then
			point = "BOTTOM"..point
			yOfs = KTF:GetBottom() and KTF:GetBottom() - ((db.qiBgrBorder and (db.bgr ~= "None" or db.border ~= "None")) and 0 or yMod)
		else
			point = "TOP"..point
			yOfs = KTF:GetTop() and KTF:GetTop() + ((db.qiBgrBorder and (db.bgr ~= "None" or db.border ~= "None")) and 0 or yMod)
		end
		if xOfs and yOfs then
			KTF.Buttons:ClearAllPoints()
			KTF.Buttons:SetPoint(point, UIParent, "BOTTOMLEFT", xOfs, yOfs)
		end
	end
end

--------------------
-- Hook functions --
--------------------

function KT:SetHooks()
	local bck_ObjectiveTracker_Update = ObjectiveTracker_Update
	ObjectiveTracker_Update = function(reason, id)
		if reason ~= OBJECTIVE_TRACKER_UPDATE_STATIC then
			_DBG("|cffffff00Update ... "..(reason and reason or ""))
		end
		bck_ObjectiveTracker_Update(reason, id)
		if db.collapsed then
			local _, numQuests = GetNumQuestLogEntries()
			local title = ("%d/%d Quests  -  %d Dailies"):format(numQuests, MAX_QUESTS, GetDailyQuestsCompleted())
			OTFHeader.Title:SetText(title)
		end
		if reason == OBJECTIVE_TRACKER_UPDATE_STATIC then
			return
		elseif reason == OBJECTIVE_TRACKER_UPDATE_TASK_ADDED then
			activeTask = id
			if db.collapsed and KT:PetTracker_IsShown() then
				return
			end
			KT:ToggleEmptyTracker(true)
		elseif activeTask then
			isInArea, isOnMap, numObjectives = GetTaskInfo(activeTask)
			if not isInArea then
				activeTask = 0
				KT:ToggleEmptyTracker()
			end
		end
		KT:SetSize()
	end

	local bck_DEFAULT_OBJECTIVE_TRACKER_MODULE_AddObjective = DEFAULT_OBJECTIVE_TRACKER_MODULE.AddObjective
	function DEFAULT_OBJECTIVE_TRACKER_MODULE:AddObjective(block, objectiveKey, text, lineType, useFullHeight, hideDash, colorStyle)
		local _, _, progress, numHave, numNeed, desc = strfind(text, "^((%d+).*/(%d+)%a?)(.*)")
		if progress and numHave ~= "0" and numNeed ~= numHave then
			text = "|cffc8c800" .. progress .. "|r" .. desc
		end
		local line = bck_DEFAULT_OBJECTIVE_TRACKER_MODULE_AddObjective(self, block, objectiveKey, text, lineType, useFullHeight, hideDash, colorStyle)
		if line.Dash then
			line.Dash:SetFont(font, db.fontSize, db.fontFlag)
			line.Dash:SetShadowColor(0, 0, 0, fontShadow)
		end
		if line.Check then
			line.Check:SetSize(db.fontSize-2.5, db.fontSize-2.5)
			line.Check:ClearAllPoints()
			line.Check:SetPoint("LEFT", -db.fontSize*0.2+(db.fontFlag == "None" and 0 or 1), 0)
		end
		return line
	end

	local bck_DEFAULT_OBJECTIVE_TRACKER_MODULE_SetStringText = DEFAULT_OBJECTIVE_TRACKER_MODULE.SetStringText
	function DEFAULT_OBJECTIVE_TRACKER_MODULE:SetStringText(fontString, text, useFullHeight, colorStyle, useHighlight)
		fontString:SetFont(font, db.fontSize, db.fontFlag)
		fontString:SetShadowColor(0, 0, 0, fontShadow)
		fontString:SetWordWrap(false)
		return bck_DEFAULT_OBJECTIVE_TRACKER_MODULE_SetStringText(self, fontString, text, useFullHeight, colorStyle, useHighlight)
	end

	hooksecurefunc(DEFAULT_OBJECTIVE_TRACKER_MODULE, "OnBlockHeaderEnter", function(self, block)
		if db.blockTooltips and (self == QUEST_TRACKER_MODULE or self == ACHIEVEMENT_TRACKER_MODULE) then
			GameTooltip:SetOwner(block, "ANCHOR_NONE")
			GameTooltip:ClearAllPoints()
			if KTF.anchorLeft then
				GameTooltip:SetPoint("TOPLEFT", block, "TOPRIGHT", -2, 0)
			else
				GameTooltip:SetPoint("TOPRIGHT", block, "TOPLEFT", -20, 0)
			end
			if self == QUEST_TRACKER_MODULE then
				GameTooltip:SetHyperlink(GetQuestLink(block.questLogIndex))
			else
				GameTooltip:SetHyperlink(GetAchievementLink(block.id))
			end
			GameTooltip:Show()
		end

		if block.fixedTag then
			local colorStyle = OBJECTIVE_TRACKER_COLOR["HeaderHighlight"]
			block.fixedTag:SetBackdropColor(colorStyle.r, colorStyle.g, colorStyle.b)
			block.fixedTag.text:SetTextColor(colorStyle.r, colorStyle.g, colorStyle.b)
		end
	end)

	hooksecurefunc(DEFAULT_OBJECTIVE_TRACKER_MODULE, "OnBlockHeaderLeave", function(self, block)
		if block.questCompleted then
			local colorStyle = OBJECTIVE_TRACKER_COLOR["Complete"]
			block.HeaderText:SetTextColor(colorStyle.r, colorStyle.g, colorStyle.b)
			block.HeaderText.colorStyle = colorStyle
		end
		if db.blockTooltips then
			GameTooltip:Hide()
		end

		if block.fixedTag then
			block.fixedTag:SetBackdropColor(1, 1, 1)
			block.fixedTag.text:SetTextColor(1, 1, 1)
		end
	end)

	hooksecurefunc(QUEST_TRACKER_MODULE, "OnBlockHeaderClick", function(self, block, mouseButton)
		GameTooltip:Hide()
	end)
	hooksecurefunc(ACHIEVEMENT_TRACKER_MODULE, "OnBlockHeaderClick", function(self, block, mouseButton)
		GameTooltip:Hide()
	end)

	local bck_ObjectiveTracker_AddBlock = ObjectiveTracker_AddBlock
	ObjectiveTracker_AddBlock = function(block, forceAdd)
		local blockAdded = bck_ObjectiveTracker_AddBlock(block, forceAdd)
		if blockAdded and (block.module == QUEST_TRACKER_MODULE or block.module == ACHIEVEMENT_TRACKER_MODULE) then
			local button = block.button
			if not button then
				button = CreateFrame("Button", nil, block)
				button:SetAllPoints()
				button:SetScript("OnEnter", function() block.module:OnBlockHeaderEnter(block) end)
				button:SetScript("OnLeave", function() block.module:OnBlockHeaderLeave(block) end)
				button:SetScript("OnClick", function(self, btn) block.module:OnBlockHeaderClick(block, btn) end)
				button:RegisterForClicks("AnyUp")
				block.button = button
			end
			button:Show()
		end
		if not db.collapsed and not blockAdded then
			OTF.BlocksFrame:SetHeight(2000)		-- slide
			OTF.BlocksFrame.contentsHeight = 2000
		end
		return blockAdded
	end

	hooksecurefunc(DEFAULT_OBJECTIVE_TRACKER_MODULE, "FreeBlock", function(self, block)
		if block.button then
			block.button:Hide()
		end
	end)

	local bck_QUEST_TRACKER_MODULE_SetBlockHeader = QUEST_TRACKER_MODULE.SetBlockHeader
	function QUEST_TRACKER_MODULE:SetBlockHeader(block, text, questLogIndex, isQuestComplete)
		local _, level, _, _, _, _, frequency, questID = GetQuestLogTitle(questLogIndex)
		local tagID, tagName = GetQuestTagInfo(questID)
		text = KT:CreateQuestTag(level, tagID, frequency) .. text
		bck_QUEST_TRACKER_MODULE_SetBlockHeader(self, block, text, questLogIndex, isQuestComplete)
		if block.questCompleted then
			local colorStyle = OBJECTIVE_TRACKER_COLOR["Complete"]
			block.HeaderText:SetTextColor(colorStyle.r, colorStyle.g, colorStyle.b)
			block.HeaderText.colorStyle = colorStyle
		end

		local link, item, charges, showItemWhenComplete = GetQuestLogSpecialItemInfo(questLogIndex)
		if item and (not isQuestComplete or showItemWhenComplete) then
			KTF.Buttons.num = KTF.Buttons.num + 1
			CreateFixedTag(block)
			local button = CreateFixedButton(block)
			if InCombatLockdown() then
				_DBG(" - STOP Create button")
				stopUpdate = true
			else
				button.questLogIndex = questLogIndex
				button.charges = charges
				button.rangeTimer = -1
				button.item = item
				button.link = link
				SetItemButtonTexture(button, item)
				SetItemButtonCount(button, charges)
				QuestObjectiveItem_UpdateCooldown(button)
				button:SetAttribute("item", link)
			end
		else
			RemoveFixedButton(block)
		end
	end

	function CreateFixedTag(block)
		block.itemButton:Hide()

		local tag = block.fixedTag
		if not tag then
			local numFreeButtons = #freeTags
			if numFreeButtons > 0 then
				tag = freeTags[numFreeButtons]
				tremove(freeTags, numFreeButtons)
				tag:SetParent(block)
			else
				tag = CreateFrame("Frame", nil, block)
				tag:SetSize(26, 26)
				tag:SetBackdrop({ bgFile = "Interface\\AddOns\\"..addonName.."\\Media\\bgr_quest-item" })
				tag.text = tag:CreateFontString(nil, "ARTWORK", "NumberFontNormal")
				tag.text:SetPoint("CENTER", 1, -1)
			end
			tag:SetPoint("TOPRIGHT")
			tag:Show()
			block.fixedTag = tag
		end

		tag:SetBackdropColor(1, 1, 1)
		tag.text:SetTextColor(1, 1, 1)
		tag.text:SetText(KTF.Buttons.num)
	end

	function CreateFixedButton(block)
		local questID = block.id
		local button = GetFixedButton(questID)
		if not button then
			if InCombatLockdown() then
				return nil
			end

			local numFreeButtons = #freeButtons
			if numFreeButtons > 0 then
				_DBG(" - USE button "..questID)
				button = freeButtons[numFreeButtons]
				tremove(freeButtons, numFreeButtons)
			else
				_DBG(" - CREATE button "..questID)
				local name = addonName.."FixedButton"..questID
				button = CreateFrame("Button", name, KTF.Buttons, "SecureActionButtonTemplate")
				button:SetSize(26, 26)

				button.icon = button:CreateTexture(name.."Icon", "BORDER")
				button.icon:SetAllPoints()

				button.Count = button:CreateFontString(name.."Count", "BORDER", "NumberFontNormal")
				button.Count:SetJustifyH("RIGHT")
				button.Count:SetPoint("BOTTOMRIGHT", button.icon, -3, 2)

				button.Cooldown = CreateFrame("Cooldown", name.."Cooldown", button, "CooldownFrameTemplate")
				button.Cooldown:SetAllPoints()

				button.HotKey = button:CreateFontString(name.."HotKey", "ARTWORK", "NumberFontNormalSmallGray")
				button.HotKey:SetSize(29, 10)
				button.HotKey:SetJustifyH("RIGHT")
				button.HotKey:SetText(RANGE_INDICATOR)
				button.HotKey:SetPoint("TOPRIGHT", button.icon, 2, -2)

				button.text = button:CreateFontString(name.."Text", "ARTWORK", "NumberFontNormalSmall")
				button.text:SetSize(29, 10)
				button.text:SetJustifyH("LEFT")
				button.text:SetPoint("TOPLEFT", button.icon, 1, -3)

				button:SetScript("OnEvent", QuestObjectiveItem_OnEvent)
				button:SetScript("OnUpdate", QuestObjectiveItem_OnUpdate)
				button:SetScript("OnShow", QuestObjectiveItem_OnShow)
				button:SetScript("OnHide", QuestObjectiveItem_OnHide)
				button:SetScript("OnEnter", QuestObjectiveItem_OnEnter)
				button:SetScript("OnLeave", QuestObjectiveItem_OnLeave)

				button:RegisterForClicks("AnyUp")
				button:SetAttribute("type","item")

				button:SetNormalTexture("Interface\\Buttons\\UI-Quickslot2")
				do local tex = button:GetNormalTexture()
					tex:ClearAllPoints()
					tex:SetPoint("CENTER")
					tex:SetSize(44, 44)
				end
				button:SetPushedTexture("Interface\\Buttons\\UI-Quickslot-Depress")
				button:SetHighlightTexture("Interface\\Buttons\\ButtonHilight-Square", "ADD")
				button:SetFrameLevel(KTF:GetFrameLevel() + 1)
				button:Hide()	-- Cooldown init

				KT:Masque_AddButton(button)
			end
			button:Show()
			fixedButtons[questID] = button
			KTF.Buttons.reanchor = true
		end
		button.block = block
		button.num = KTF.Buttons.num
		button.text:SetText(KTF.Buttons.num)
		button:SetAlpha(1)
		if KTF.ActiveButton.questID == questID then
			KT:UpdateActiveButton(questID)
		end
		return button
	end

	function RemoveFixedButton(block)
		if block then
			local tag = block.fixedTag
			if tag then
				tinsert(freeTags, tag)
				block.fixedTag = nil
				tag:Hide()
			end
			local questID = block.id
			local button = GetFixedButton(questID)
			if button then
				button:SetAlpha(0)
				if InCombatLockdown() then
					_DBG(" - STOP Remove button")
					stopUpdate = true
				else
					_DBG(" - REMOVE button "..questID)
					tinsert(freeButtons, button)
					fixedButtons[questID] = nil
					button:Hide()
					KTF.Buttons.reanchor = true
				end
				KTF.ActiveButton.text:SetText("")
			end
			if KTF.Buttons.timerID == questID then
				KTF.Buttons.timerID = nil
			end
		else
			for questID, button in pairs(fixedButtons) do
				_DBG(" - REMOVE button "..questID)
				tinsert(freeButtons, button)
				fixedButtons[questID] = nil
				button:Hide()
			end
		end
	end

	function GetFixedButton(questID)
		if fixedButtons[questID] then
			return fixedButtons[questID]
		else
			return nil
		end
	end

	function QuestObjectiveItem_OnEnter(self)
		if KTF.anchorLeft then
			GameTooltip:SetOwner(self, "ANCHOR_RIGHT", 2)
		else
			GameTooltip:SetOwner(self, "ANCHOR_LEFT", -2)
		end
		GameTooltip:SetQuestLogSpecialItem(self.questLogIndex)

		local block = self.block
		local colorStyle = OBJECTIVE_TRACKER_COLOR["HeaderHighlight"]
		block.HeaderText:SetTextColor(colorStyle.r, colorStyle.g, colorStyle.b)
		block.HeaderText.colorStyle = colorStyle

		if block.fixedTag then
			block.fixedTag:SetBackdropColor(colorStyle.r, colorStyle.g, colorStyle.b)
			block.fixedTag.text:SetTextColor(colorStyle.r, colorStyle.g, colorStyle.b)
		end
	end

	function QuestObjectiveItem_OnLeave(self)
		GameTooltip:Hide()

		local colorStyle
		local block = self.block
		if block.questCompleted then
			colorStyle = OBJECTIVE_TRACKER_COLOR["Complete"]
		else
			colorStyle = OBJECTIVE_TRACKER_COLOR["Header"]
		end
		block.HeaderText:SetTextColor(colorStyle.r, colorStyle.g, colorStyle.b)
		block.HeaderText.colorStyle = colorStyle

		if block.fixedTag then
			block.fixedTag:SetBackdropColor(1, 1, 1)
			block.fixedTag.text:SetTextColor(1, 1, 1)
		end
	end

	function QuestObjectiveItem_OnUpdate(self, elapsed)
		local rangeTimer = self.rangeTimer
		if rangeTimer then
			rangeTimer = rangeTimer - elapsed
			if rangeTimer <= 0 then
				local link, item, charges, showItemWhenComplete = GetQuestLogSpecialItemInfo(self.questLogIndex)
				if not charges or charges ~= self.charges then
					ObjectiveTracker_Update(OBJECTIVE_TRACKER_UPDATE_MODULE_QUEST)
					return
				end
				local hotkey = self.HotKey
				local valid = IsQuestLogSpecialItemInRange(self.questLogIndex)
				if hotkey:GetText() == RANGE_INDICATOR then
					if valid == 0 then
						hotkey:Show()
						hotkey:SetVertexColor(1.0, 0.1, 0.1)
					elseif valid == 1 then
						hotkey:Show()
						hotkey:SetVertexColor(0.6, 0.6, 0.6)
					else
						hotkey:Hide()
					end
				else
					if valid == 0 then
						hotkey:SetVertexColor(1.0, 0.1, 0.1)
					else
						hotkey:SetVertexColor(0.6, 0.6, 0.6)
					end
				end
				rangeTimer = TOOLTIP_UPDATE_TIME
			end
			self.rangeTimer = rangeTimer
		end

		if db.qiActiveButton and not InCombatLockdown() then
			if KTF.Buttons.timerID then
				if KTF.Buttons.timerID ~= self.block.id then
					return
				end
			else
				KTF.Buttons.timerID = self.block.id
			end
			if KTF.Buttons.timer > 50 then
				KT:UpdateActiveButton()
			else
				--_DBG("... "..KTF.Buttons.timer.." ... "..self.block.id)
				KTF.Buttons.timer = KTF.Buttons.timer + TOOLTIP_UPDATE_TIME
			end
		end
	end

	local bck_QUEST_TRACKER_MODULE_OnFreeBlock = QUEST_TRACKER_MODULE.OnFreeBlock
	function QUEST_TRACKER_MODULE:OnFreeBlock(block)
		RemoveFixedButton(block)
		bck_QUEST_TRACKER_MODULE_OnFreeBlock(self, block)
	end

	local bck_QUEST_TRACKER_MODULE_Update = QUEST_TRACKER_MODULE.Update
	function QUEST_TRACKER_MODULE:Update()
		KTF.Buttons.num = 0
		bck_QUEST_TRACKER_MODULE_Update()
	end

	hooksecurefunc(QUEST_TRACKER_MODULE, "EndLayout", function(self, isStaticReanchor)
		if not InCombatLockdown() and KTF.Buttons.reanchor then
			_DBG(" - REANCHOR buttons")
			local questID, block, yOfs
			local numButtons = 0
			local contentsHeight = 0
			for i=1, GetNumQuestWatches() do
				questID, _ = GetQuestWatchInfo(i)
				block = QUEST_TRACKER_MODULE.usedBlocks[questID]
				if block and block.itemButton then
					numButtons = numButtons + 1
					yOfs = -(contentsHeight + 7)
					contentsHeight = contentsHeight + 26 + 3
					fixedButtons[questID]:SetPoint("TOP", 0, yOfs)
				end
			end
			KTF.Buttons:SetHeight(contentsHeight + 7 + 4)
			KTF.Buttons.height = contentsHeight
			KTF.Buttons.reanchor = false

			KT:UpdateActiveButton()
		end
		if db.collapsed or KTF.Buttons.height == 0 then
			KTF.Buttons:SetAlpha(0)
		else
			KTF.Buttons:SetAlpha(1)
		end
	end)

	local bck_BONUS_OBJECTIVE_TRACKER_MODULE_AddProgressBar = BONUS_OBJECTIVE_TRACKER_MODULE.AddProgressBar
	function BONUS_OBJECTIVE_TRACKER_MODULE:AddProgressBar(block, line, questID, finished)
		local progressBar = bck_BONUS_OBJECTIVE_TRACKER_MODULE_AddProgressBar(self, block, line, questID, finished)
		progressBar.Bar.Label:SetPoint("CENTER")
		progressBar.Bar.Label:SetFont(font, 13, db.fontFlag)
		progressBar.Bar.Label:SetShadowColor(0, 0, 0, fontShadow)
		progressBar.Bar.Label:Show()
		progressBar.Bar:SetScript("OnEnter", function() BonusObjectiveTracker_OnBlockEnter(block) end)
		progressBar.Bar:SetScript("OnLeave", function() BonusObjectiveTracker_OnBlockLeave(block) end)
		return progressBar
	end

	hooksecurefunc("BonusObjectiveTracker_ShowRewardsTooltip", function(block)
		if KTF.anchorLeft then
			GameTooltip:ClearAllPoints()
			GameTooltip:SetPoint("TOPLEFT", block, "TOPRIGHT", 10, 0)
		end
	end)

	function ScenarioObjectiveStageBlock_OnEnter2(self)
		ScenarioObjectiveStageBlock_OnEnter(self)
		if KTF.anchorLeft then
			GameTooltip:ClearAllPoints()
			GameTooltip:SetPoint("LEFT", self, "RIGHT", 42, 0)
		end
	end
	ScenarioStageBlock:SetScript("OnEnter", ScenarioObjectiveStageBlock_OnEnter2)

	function ObjectiveTracker_Collapse()
		_DBG("--------------------------------")
		_DBG("COLLAPSE")
		OTF.collapsed = true
		self.db.profile.collapsed = OTF.collapsed
		OTF.BlocksFrame:Hide()
		OTFButton:GetNormalTexture():SetTexCoord(0, 0.5, 0, 0.25)
		OTFHeader.Title:Show()
		CloseDropDownMenus()
		animTask = false		
	end

	function ObjectiveTracker_Expand()
		_DBG("--------------------------------")
		_DBG("EXPAND")
		OTF.collapsed = nil
		self.db.profile.collapsed = OTF.collapsed
		OTF.BlocksFrame:Show()
		OTFButton:GetNormalTexture():SetTexCoord(0, 0.5, 0.25, 0.5)
		OTFHeader.Title:Hide()
	end

	hooksecurefunc("BonusObjectiveTracker_OnBlockAnimOutFinished", function(self)
		animTask = false
		if KT:PetTracker_IsShown() then
			local yOfs = PetTracker.Objectives:GetUsedHeight()
			PetTracker.Objectives:SetPoint("TOPLEFT", OTF, -10, -yOfs)
		end
	end)

	hooksecurefunc("BonusObjectiveTracker_SetBlockState", function(block, state, force)
		if state == "PRESENT" then
			_DBG(" - "..state)
			activeTask = block.id
		elseif state == "LEAVING" and activeTask > 0 then
			_DBG(" - "..state)
			animTask = true
			activeTask = 0
			if KT:IsTrackerEmpty(true) then
				OTF.BlocksFrame.contentsHeight = OTF.BlocksFrame.contentsHeight + paddingBottom
			end
		end		
	end)

	hooksecurefunc("ExtraActionBar_Update", function()
		if KTF.ActiveButton.questID then
			local bar = ExtraActionBarFrame
			bar.outro:Stop()
			bar:SetAlpha(1)
		end
	end)

	function QuestActiveButton_UpdateHotkey()
		local key = GetBindingKey("EXTRAACTIONBUTTON1")
		if db.keyBindMinimize == key then
			SetOverrideBinding(KTF, false, db.keyBindMinimize, nil)
			db.keyBindMinimize = ""
		end
		if db.qiActiveButton then
			local abutton = KTF.ActiveButton
			local hotkey = abutton.HotKey
			local text = GetBindingText(key, 1)
			ClearOverrideBindings(abutton)
			if key then
				SetOverrideBindingClick(abutton, false, key, abutton:GetName())
			end
			if text == "" then
				hotkey:SetText(RANGE_INDICATOR)
				hotkey:Hide()
			else
				hotkey:SetText(text)
				hotkey:Show()
			end
		end
	end
end

function KT:UpdateActiveButton(id)
	if not db.qiActiveButton then
		return
	end

	local button
	local abutton = KTF.ActiveButton
	local autoShowTooltip = false

	if id then
		button = GetFixedButton(id)
		if GameTooltip:IsShown() and GameTooltip:GetOwner() == abutton then
			QuestObjectiveItem_OnLeave(abutton)
			autoShowTooltip = true
		end
		abutton.block = button.block
		abutton.text:SetText(button.num)
		if autoShowTooltip then
			QuestObjectiveItem_OnEnter(abutton)
		end
		return
	end

	if InCombatLockdown() then
		return
	end

	local closestQuestID
	local minDistSqr = 22500
	local bar = ExtraActionBarFrame

	for questID, button in pairs(fixedButtons) do
		if QuestHasPOIInfo(questID) then
			local distSqr, onContinent = GetDistanceSqToQuest(button.questLogIndex)
			if onContinent and distSqr <= minDistSqr then
				minDistSqr = distSqr
				closestQuestID = questID
			end
		end
	end

	if closestQuestID then
		button = GetFixedButton(closestQuestID)
		if abutton.questID ~= closestQuestID or not bar:IsShown() then
			if GameTooltip:IsShown() and GameTooltip:GetOwner() == abutton then
				QuestObjectiveItem_OnLeave(abutton)
				autoShowTooltip = true
			end

			bar:Show()
			bar.button:SetAlpha(0)

			abutton:Show()
			abutton.block = button.block
			abutton.questID = closestQuestID
			abutton.questLogIndex = button.questLogIndex
			abutton.charges = button.charges
			abutton.rangeTimer = button.rangeTimer
			SetItemButtonTexture(abutton, button.item)
			SetItemButtonCount(abutton, button.charges)
			QuestObjectiveItem_UpdateCooldown(abutton)
			abutton:SetAttribute("item", button.link)
			QuestActiveButton_UpdateHotkey()

			UIParent_ManageFramePositions()
			bar.outro:Stop()
			bar:SetAlpha(1)

			if autoShowTooltip then
				QuestObjectiveItem_OnEnter(abutton)
			end
		end
		abutton.text:SetText(button.num)
	elseif bar:IsShown() then
		bar.button:SetAlpha(1)
		if not HasExtraActionBar() then
			bar.intro:Stop()
			bar:SetAlpha(0)
			bar:Hide()
		end

		abutton:Hide()
		ClearOverrideBindings(abutton)
	end
	KTF.Buttons.timer = 0
end

function KT:ToggleTrackAllQuests()
	local firstQuestID
	local numWatches = GetNumQuestWatches()
	local numEntries, numQuests = GetNumQuestLogEntries()
	if numWatches > 0 then
		for questLogIndex=1, numEntries do
			local _, _, _, isHeader = GetQuestLogTitle(questLogIndex)
			if not isHeader then
				RemoveQuestWatch(questLogIndex)
			end
		end
	else
		for questLogIndex=1, numEntries do
			local _, _, _, isHeader, _, _, _, questID = GetQuestLogTitle(questLogIndex)
			if not isHeader and not IsQuestWatched(questLogIndex) then
				AddQuestWatch(questLogIndex, true)
				firstQuestID = questID
			end
		end
		QuestSuperTracking_OnQuestTracked(firstQuestID)
		ObjectiveTracker_Update()	-- preventive
	end
end

function KT:CreateQuestTag(level, questTag, frequency)
	local tag = ""
	
	if level == -1 then
		level = "*"
	else
		level = tostring(level)
	end
	
	if questTag == ELITE then
		tag = "+"
	elseif questTag == QUEST_TAG_GROUP then
		tag = "g"
	elseif questTag == QUEST_TAG_PVP then
		tag = "pvp"
	elseif questTag == QUEST_TAG_DUNGEON then
		tag = "d"
	elseif questTag == QUEST_TAG_HEROIC then
		tag = "hc"
	elseif questTag == QUEST_TAG_RAID then
		tag = "r"
	elseif questTag == QUEST_TAG_RAID10 then
		tag = "r10"
	elseif questTag == QUEST_TAG_RAID25 then
		tag = "r25"
	elseif questTag == QUEST_TAG_SCENARIO then
		tag = "s"
	elseif questTag == QUEST_TAG_ACCOUNT then
		tag = "a"
	elseif questTag == QUEST_TAG_LEGENDARY then
		tag = "leg"
	end
	
	if frequency == 2 then
		tag = tag.."!"		-- daily quest
	elseif frequency == 3 then
		tag = tag.."!!"	-- weekly quest
	end
	
	if tag ~= "" then
		tag = ("|cff00b3ff%s|r"):format(tag)
	end
	
	tag = ("[%s%s] "):format(level, tag)
	return tag
end

function KT:SetHeader(obj)
	obj.Text:SetFont(font, db.fontSize+2, db.fontFlag)
	obj.Text:SetShadowColor(0, 0, 0, fontShadow)
	obj.Text:SetPoint("LEFT", 4, 0)
	obj.animateReason = 0
end

function KT:ToggleEmptyTracker(added)
	local alpha, mouse = 1, true
	if self:IsTrackerEmpty() then
		if not db.collapsed then
			ObjectiveTracker_MinimizeButton_OnClick()										
		end
		OTFButton:GetNormalTexture():SetTexCoord(0, 0.5, 0.5, 0.75)
		if db.hideEmptyTracker then
			alpha = 0
			mouse = false
		end
	else
		if db.collapsed then
			if added then
				ObjectiveTracker_MinimizeButton_OnClick()
			else
				OTFButton:GetNormalTexture():SetTexCoord(0, 0.5, 0, 0.25)
			end
		else
			OTFButton:GetNormalTexture():SetTexCoord(0, 0.5, 0.25, 0.5)
		end
	end
	KTF:SetAlpha(alpha)
	OTFHeader.MinimizeButton:EnableMouse(mouse)
	if db.hbOtherButtons then
		OTFHeader.QuestLogButton:EnableMouse(mouse)
		OTFHeader.AchievementsButton:EnableMouse(mouse)
	end
end

function KT:IsTrackerEmpty(noaddon)
	local result = (GetNumQuestWatches() == 0 and
			GetNumAutoQuestPopUps() == 0 and
			GetNumTrackedAchievements() == 0 and
			activeTask == 0 and
			not C_Scenario.IsInScenario())
	if not noaddon then
		result = (result and
			  not self:PetTracker_IsShown())
	end
	return result
end

--------------------------------
-- Supported addons functions --
--------------------------------

-- PetTracker
function KT:PetTracker_SetSupport()
	IsPetTracker = (db.addonPetTracker and IsAddOnLoaded("PetTracker"))
	self:CheckAddOnCompat("PetTracker", "6.0.17")
end

function KT:PetTracker_IsShown()
	return (IsPetTracker and
		not PetTracker.Sets.HideTracker and
		PetTracker.Objectives:IsShown())
end

function KT:PetTracket_SetHooks()
	if IsPetTracker then
		PetTracker.Objectives:SetParent(OTF)
		PetTracker.Objectives.Header:SetParent(OTF.BlocksFrame)
		PetTracker.Objectives.Anchor:SetParent(OTF.BlocksFrame)
		PetTracker.Objectives.Anchor.text:SetPoint("CENTER", 0, -2)
		
		local bck_PetTracker_Objectives_TrackingChanged = PetTracker.Objectives.TrackingChanged
		function PetTracker.Objectives:TrackingChanged()
			if not PetTracker.Sets.HideTracker then
				_DBG(" - PT Update ... "..self.maxEntries)
				self.maxEntries = 100
				bck_PetTracker_Objectives_TrackingChanged(self)
				self.maxEntries = self:Count()
				self.maxEntries = self.maxEntries > 0 and self.maxEntries + 1 or 0
				_DBG("    maxEntries = "..self.maxEntries)
				self.Header:SetShown(self.Anchor:IsShown())
			end
		end
		
		-- Replace original function!
		function PetTracker.Objectives:GetUsedHeight()
			local mod = (KT:IsTrackerEmpty(true) and 0 or paddingBottom)
			-- PopUp Quest
			if AUTO_QUEST_POPUP_TRACKER_MODULE.firstBlock then
				mod = mod + (GetNumAutoQuestPopUps() * 68)
			end
			-- Bonus objectives - problem with order of events
			if animTask or activeTask > 0 then
				mod = mod + BONUS_OBJECTIVE_TRACKER_MODULE.blockPadding
			end
			return round(OTF.BlocksFrame.contentsHeight) + mod
		end
		
		-- Replace original function!
		function PetTracker.Tracker:AddSpecie(specie, quality, level)
			local Journal = PetTracker.Journal
			local sourceIcon = Journal:GetSourceIcon(specie)
			if sourceIcon then
				-- original content
				local name, icon = Journal:GetInfo(specie)
				local r,g,b = self:GetColor(quality)
				local line = self:NewLine()
				line.Text:SetText(name .. (level > 0 and format(' (%s)', level) or ''))
				line.Text:SetWidth(self.Anchor:GetWidth())
				line.SubIcon:SetTexture(sourceIcon)
				line.Icon:SetTexture(icon)
				line:SetScript('OnClick', function()
					Journal:Display(specie)
				end)
				line:SetScript('OnEnter', function()
					line.Text:SetTextColor(r, g, b)
				end)
				line:SetScript('OnLeave', function()
					line.Text:SetTextColor(r-.2, g-.2, b-.2)
				end)
				line:GetScript('OnLeave')(line)
				-- added content
				line.Text:SetFont(font, db.fontSize, db.fontFlag)
				line.Text:SetShadowColor(0, 0, 0, fontShadow)
				line.Text:ClearAllPoints()
				line.Text:SetPoint("LEFT", line.Icon, "RIGHT", 10, 0)
				line:SetParent(OTF.BlocksFrame)
			end
		end
		
		hooksecurefunc(PetTracker, "ForAllModules", function(self, event, ...)
			if event == "TrackingChanged" then
				KT:SetSize()
				KT:ToggleEmptyTracker()
			end				
		end)
		
		-- Sushi Lib - hack - revert back DropDownMenu
		-- Replace original functions!
		if SushiDropFrame then
			local Drop = CreateFrame("Frame", "DropdownFrameFix", nil, "UIDropDownMenuTemplate")
			function Drop:AddLine(data)
				UIDropDownMenu_AddButton(data)
			end
			function SushiDropFrame:Toggle(...)
				local n = select("#", ...)
				local anchor = select(n < 4 and 1 or 2, ...)
				Drop.initialize = select(n, ...)
				Drop.displayMode = "MENU"
				ToggleDropDownMenu(1, nil, Drop, anchor, 0, -5)
			end
			SushiDropFrame.CloseAll = function() end
		end
	elseif PetTracker then
		-- Disable tracking
		-- Replace original functions!
		PetTracker.Tracker.Toggle = function() end
		PetTracker.Objectives.GetUsedHeight = function() return 0 end
	end
end

-- Masque
function KT:Masque_SetSupport()
	self:CheckAddOnCompat("Masque", "6.0.0")
end

function KT:Masque_AddButton(button)
	if db.addonMasque and MSQ then
		MSQ:Group(KT.title):AddButton(button)
		if button.Style then
			button.Style:SetAlpha(0)
		end
	end
end

-- ElvUI (preventive)
function KT:ElvUI_SetSupport()
	if IsAddOnLoaded("ElvUI") then
		local E = unpack(_G.ElvUI)
		E.Blizzard.MoveObjectiveFrame = function() end
		hooksecurefunc(E, "ToggleConfig", function(self)
			local options = self.Options.args.general.args.objectiveFrame
			options.disabled = true
			options.args.KTwarning = {
				name = "|cffff7f00Addon "..KT.title.." is active.\n",
				width = "full",
				type = "description",
				order = 0,
			}
		end)
	end
end

-- Tukui (preventive)
function KT:Tukui_SetSupport()
	if IsAddOnLoaded("Tukui") then
		local T = unpack(_G.Tukui)
		T.Miscellaneous.ObjectiveTracker.Enable = function() end
	end
end