
--local qgc = QuestGuru
local addonName, addon = ...
local QGT = LibStub("AceAddon-3.0"):GetAddon(addonName)
QGT.title = GetAddOnMetadata(addonName, "Title")
QGT.version = GetAddOnMetadata(addonName, "Version")

local LSM = LibStub("LibSharedMedia-3.0")
local _DBG = function(...) if _DBG then _DBG("QGT", ...) end end

local _G = _G

local floor = math.floor
local format = string.format
local hooksecurefunc = hooksecurefunc
local pairs = pairs
local strfind = string.find
local tinsert = table.insert
local tremove = table.remove
local unpack = unpack
local round = function(n) return floor(n + 0.5) end

local CreateFrame = CreateFrame
local InCombatLockdown = InCombatLockdown
local UIParent = UIParent
local GameTooltip = GameTooltip

local trackerWidth = 280
local paddingBottom = 16
	
local fixedButtons = {}
local freeTags = {}
local freeButtons = {}
	
local stopUpdate = false
local activeTask = 0
local animTask = false

local db, hbColor, font, fontShadow 

-- Main frame
local QGTF = CreateFrame("Frame", addonName.."Frame", UIParent)
QGT.frame = QGTF

-- Blizzard OTF objects
local OTF = ObjectiveTrackerFrame
local OTFHeader = OTF.HeaderMenu
local OTFButton = OTF.HeaderMenu.MinimizeButton

OBJECTIVE_TRACKER_COLOR = {
	["Normal"] = { r = 0.8, g = 0.8, b = 0.8 },
	["NormalHighlight"] = { r = HIGHLIGHT_FONT_COLOR.r, g = HIGHLIGHT_FONT_COLOR.g, b = HIGHLIGHT_FONT_COLOR.b },
	["Failed"] = { r = DIM_RED_FONT_COLOR.r, g = DIM_RED_FONT_COLOR.g, b = DIM_RED_FONT_COLOR.b },
	["FailedHighlight"] = { r = RED_FONT_COLOR.r, g = RED_FONT_COLOR.g, b = RED_FONT_COLOR.b },
	["Header"] = { r = 1, g = 0.5, b = 0 },		-- orange
	["HeaderHighlight"] = { r = NORMAL_FONT_COLOR.r, g = NORMAL_FONT_COLOR.g, b = NORMAL_FONT_COLOR.b },
	["Complete"] = { r = 0, g = 1, b = 0 },		-- green
}

--------------------
-- Load functions --
--------------------

function QGT:OnInitialize()
	-- Get character data
	local _, class = UnitClass("player")
	QGT.classColor = RAID_CLASS_COLORS[class]
	
	-- Setup Options
	QGT:SetupOptions()
	db = QGT.db.profile
end

function QGT:OnEnable()
	QGT.screenWidth = floor(GetScreenWidth())
	QGT.screenHeight = floor(GetScreenHeight())
	
	OTFHeader:Show()
	OTFHeader.Hide = function() end
	OTFHeader.SetShown = function() end
	OTFHeader.Title:ClearAllPoints()
	OTFHeader.Title:SetPoint("LEFT", OTF, "LEFT", -20, 0)
	
	QGTF:RegisterEvent("PLAYER_ENTERING_WORLD")
	QGTF:SetScript("OnEvent", function()
		QGT:SetHooks()
		QGT:SetFrames()
		QGT:Init()
		
		if db.collapsed then
			ObjectiveTracker_MinimizeButton_OnClick()
		end
	end)
end

--------------------
-- Init functions --
--------------------

function QGT:Init()
	db = QGT.db.profile
	
	if db.keyBindMinimize ~= "" then
		SetOverrideBindingClick(QGTF, false, db.keyBindMinimize, OTFButton:GetName())
	end
	
	QGT:MoveTracker()
	QGT:SetBackground()
	QGT:SetText()
	ObjectiveTracker_Update()
end

function QGT:SetFrames()
	-- Main frame
	QGTF:SetWidth(trackerWidth)
	QGTF:SetFrameStrata(db.frameStrata)
	QGTF:SetFrameLevel(QGTF:GetFrameLevel() + 25)
	
	if not db.collapseInInstance then
		QGTF:UnregisterEvent("PLAYER_ENTERING_WORLD")
	end
	QGTF:RegisterEvent("QUEST_WATCH_LIST_CHANGED")
	QGTF:RegisterEvent("TRACKED_ACHIEVEMENT_LIST_CHANGED")
	QGTF:RegisterEvent("QUEST_ACCEPTED")
	QGTF:RegisterEvent("QUEST_AUTOCOMPLETE")
	QGTF:RegisterEvent("PLAYER_REGEN_ENABLED")
	QGTF:RegisterEvent("ZONE_CHANGED")
	QGTF:RegisterEvent("ZONE_CHANGED_NEW_AREA")
	QGTF:RegisterEvent("QUEST_POI_UPDATE")
	QGTF:RegisterEvent("WORLD_MAP_UPDATE")
	QGTF:RegisterEvent("UPDATE_BINDINGS")
--	QGTF:RegisterEvent("QUEST_LOG_UPDATE")
--	QGTF:RegisterEvent("SCENARIO_UPDATE")
--	QGTF:RegisterEvent("SCENARIO_CRITERIA_UPDATE")
--	QGTF:RegisterEvent("TRACKED_ACHIEVEMENT_UPDATE")
--	QGTF:RegisterEvent("VARIABLES_LOADED")
--	QGTF:RegisterEvent("QUEST_TURNED_IN")
--	QGTF:RegisterEvent("PET_JOURNAL_LIST_UPDATE")

	QGTF:SetScript("OnEvent", function(_, event, ...)
		_DBG("Event - "..event)
		if event == "PLAYER_ENTERING_WORLD" then
			if IsInInstance() and not db.collapsed then
				ObjectiveTracker_MinimizeButton_OnClick()
			end
		elseif event == "QUEST_WATCH_LIST_CHANGED" or event == "TRACKED_ACHIEVEMENT_LIST_CHANGED" then
			local questID, added = ...
			QGT:ToggleEmptyTracker(added)
		elseif event == "QUEST_ACCEPTED" then
			local questLogIndex, questID = ...
		elseif event == "QUEST_AUTOCOMPLETE" then
			if db.collapsed then
				ObjectiveTracker_MinimizeButton_OnClick()
			end
		elseif event == "PLAYER_REGEN_ENABLED" and stopUpdate then
			stopUpdate = false
			RemoveFixedButton()
			if not db.collapsed then
				ObjectiveTracker_Update(OBJECTIVE_TRACKER_UPDATE_QUEST)
			end
		elseif (event == "ZONE_CHANGED" or event == "ZONE_CHANGED_NEW_AREA") and QGTF.Buttons.num > 0 then
			QGTF.Buttons.reanchor = not InCombatLockdown()
		elseif event == "UPDATE_BINDINGS" then
			QuestActiveButton_UpdateHotkey()
		end
		if event == "QUEST_WATCH_LIST_CHANGED" or event == "QUEST_POI_UPDATE" or event == "ZONE_CHANGED" or event == "WORLD_MAP_UPDATE" then
			QGT:UpdateActiveButton()
		end
	end)
	
	local button
	-- Minimize button
	OTFButton:Hide()
	button = CreateFrame("Button", addonName.."MinimizeButton", QGTF)
	button:SetSize(16, 16)
	button:SetPoint("TOPRIGHT", -10, -8)
	button:SetFrameLevel(QGTF:GetFrameLevel() + 10)
	button:SetNormalTexture("Interface\\AddOns\\QuestGuru\\Graphics\\btns_header")
	button:GetNormalTexture():SetTexCoord(0, 0.5, 0.25, 0.5)
		
	button:RegisterForClicks("AnyDown")
	button:SetScript("OnClick", function(QGT, btn)
		if IsAltKeyDown() then
			QGT:OpenOptions()
		elseif IsControlKeyDown() then
			QGT:ToggleTrackAllQuests()
		elseif not QGT:IsTrackerEmpty() then
			ObjectiveTracker_MinimizeButton_OnClick()
		end
	end)
	button:SetScript("OnEnter", function(QGT)
		QGT:GetNormalTexture():SetVertexColor(1, 1, 1)
		GameTooltip:SetOwner(QGT, "ANCHOR_RIGHT")
		local title = db.keyBindMinimize ~= "" and QGT.title.." "..NORMAL_FONT_COLOR_CODE.."("..db.keyBindMinimize..")|r" or QGT.title
		GameTooltip:AddLine(title, 1, 1, 1)
		GameTooltip:AddLine("Ctrl+Click - track/untrack all quests", 0.5, 0.5, 0.5)
		GameTooltip:AddLine("Alt+Click - addon Options", 0.5, 0.5, 0.5)
		GameTooltip:Show()
	end)
	button:SetScript("OnLeave", function(QGT)
		QGT:GetNormalTexture():SetVertexColor(hbColor.r, hbColor.g, hbColor.b)
		GameTooltip:Hide()
	end)
	OTFButton = button
	OTF.HeaderMenu.MinimizeButton = button
	
	if db.hbOtherButtons then
		-- Quest Log button
		button = CreateFrame("Button", addonName.."QuestLogButton", QGTF)
		button:SetSize(16, 16)
		button:SetPoint("TOPRIGHT", -50, -8)
		button:SetFrameLevel(QGTF:GetFrameLevel() + 10)
		button:SetNormalTexture("Interface\\AddOns\\QuestGuru\\Graphics\\btns_header")
		button:GetNormalTexture():SetTexCoord(0.5, 1, 0, 0.25)
		
		button:RegisterForClicks("AnyDown")
		button:SetScript("OnClick", function(QGT, btn)
			ToggleQuestLog()
		end)
		button:SetScript("OnEnter", function(QGT)
			QGT:GetNormalTexture():SetVertexColor(1, 1, 1)
			GameTooltip:SetOwner(QGT, "ANCHOR_RIGHT")
			GameTooltip:AddLine(QuestLogMicroButton.tooltipText, 1, 1, 1)
			GameTooltip:Show()
		end)
		button:SetScript("OnLeave", function(QGT)
			QGT:GetNormalTexture():SetVertexColor(hbColor.r, hbColor.g, hbColor.b)
			GameTooltip:Hide()
		end)
		OTFHeader.QuestLogButton = button
		
		-- Achievements button
		button = CreateFrame("Button", addonName.."AchievementsButton", QGTF)
		button:SetSize(16, 16)
		button:SetPoint("TOPRIGHT", -30, -8)
		button:SetFrameLevel(QGTF:GetFrameLevel() + 10)
		button:SetNormalTexture("Interface\\AddOns\\QuestGuru\\Graphics\\btns_header")
		button:GetNormalTexture():SetTexCoord(0.5, 1, 0.25, 0.5)
		
		button:RegisterForClicks("AnyDown")
		button:SetScript("OnClick", function(QGT, btn)
			ToggleAchievementFrame()
		end)
		button:SetScript("OnEnter", function(QGT)
			QGT:GetNormalTexture():SetVertexColor(1, 1, 1)
			GameTooltip:SetOwner(QGT, "ANCHOR_RIGHT")
			GameTooltip:AddLine(AchievementMicroButton.tooltipText, 1, 1, 1)
			GameTooltip:Show()
		end)
		button:SetScript("OnLeave", function(QGT)
			QGT:GetNormalTexture():SetVertexColor(hbColor.r, hbColor.g, hbColor.b)
			GameTooltip:Hide()
		end)
		OTFHeader.AchievementsButton = button
	end
	
	-- Scroll frame
	local Scroll = CreateFrame("ScrollFrame", addonName.."Scroll", QGTF)
	Scroll:SetPoint("TOPLEFT", 34, -5)
	Scroll:SetPoint("BOTTOMRIGHT", -4, 5)
	--Scroll:SetHitRectInsets(-30, 0, 0, 1)
	Scroll:EnableMouseWheel(true)
	Scroll.step = 20
	Scroll.value = 0
	Scroll:SetScript("OnMouseWheel", function(QGT, delta)
		if not db.collapsed and OTF.height > db.maxHeight then	-- floats ??? +11
			if delta < 0 then
				QGT.value = (QGT.value+QGT.step < OTF.height-db.maxHeight) and QGT.value + QGT.step or OTF.height - db.maxHeight
			else
				QGT.value = (QGT.value-QGT.step > 0) and QGT.value - QGT.step or 0
			end
			QGT:SetVerticalScroll(QGT.value)
			if QGT.value > 0 and QGT.value < OTF.height-db.maxHeight then
				CloseDropDownMenus()
			end				
			_DBG("SCROLL ... "..QGT.value.." ... "..OTF.height.." - "..db.maxHeight)
		end
	end)
	Scroll:SetScript("OnScrollRangeChanged", function(QGT, x, y)
		QGT:MoveButtons()
	end)
	QGTF.Scroll = Scroll

	-- Blizzard frame
	OTF:SetParent(QGTF.Scroll)
	OTF:SetClampedToScreen(false)
	OTF:SetMovable(false)
	OTF.SetMovable = function() end
	OTF:ClearAllPoints()
	OTF.ClearAllPoints = function() end
	OTF:SetPoint("TOPLEFT")
	OTF:SetPoint("BOTTOMRIGHT")
	OTF.SetPoint = function() end
	QGTF.Scroll:SetScrollChild(OTF)
	
	-- Buttons frame
	local Buttons = CreateFrame("Frame", addonName.."Buttons", UIParent)
	Buttons:SetSize(40, 40)
	Buttons:SetPoint("TOPLEFT", 0, 0)
	Buttons:SetFrameStrata(db.frameStrata)
	Buttons:SetFrameLevel(QGTF:GetFrameLevel() - 1)
	Buttons:SetAlpha(0)
	
	Buttons.num = 0
	Buttons.height = 0
	Buttons.timer = 0
	Buttons.timerID = nil
	Buttons.reanchor = false
	QGTF.Buttons = Buttons
	
	-- Active (closest) button frame
	local name = addonName.."ActiveButton"
	local abutton = CreateFrame("Button", name, ExtraActionBarFrame, "SecureActionButtonTemplate")
	abutton:SetSize(52, 52)
	abutton:SetPoint("CENTER", 0, 0.5)
	
	abutton.icon = abutton:CreateTexture(name.."Icon", "BACKGROUND")
	abutton.icon:SetPoint("TOPLEFT", 0, -1)
	abutton.icon:SetPoint("BOTTOMRIGHT", 0, -1)
	
	abutton.Style = abutton:CreateTexture(name.."Style", "OVERLAY")
	abutton.Style:SetSize(256, 128)
	abutton.Style:SetPoint("CENTER", -2, 0)
	abutton.Style:SetTexture("Interface\\ExtraButton\\Default")
	
	abutton.Count = abutton:CreateFontString(name.."Count", "OVERLAY", "NumberFontNormal")
	abutton.Count:SetJustifyH("RIGHT")
	abutton.Count:SetPoint("BOTTOMRIGHT", abutton.icon, -2, 2)
	
	abutton.Cooldown = CreateFrame("Cooldown", name.."Cooldown", abutton, "CooldownFrameTemplate")
	abutton.Cooldown:ClearAllPoints()
	abutton.Cooldown:SetPoint("TOPLEFT", 4, -4)
	abutton.Cooldown:SetPoint("BOTTOMRIGHT", -3, 2)
	
	abutton.HotKey = abutton:CreateFontString(name.."HotKey", "ARTWORK", "NumberFontNormalSmallGray")
	abutton.HotKey:SetSize(30, 10)
	abutton.HotKey:SetJustifyH("RIGHT")
	abutton.HotKey:SetText(RANGE_INDICATOR)
	abutton.HotKey:SetPoint("TOPRIGHT", abutton.icon, -2, -7)
	
	abutton.text = abutton:CreateFontString(name.."Text", "ARTWORK", "NumberFontNormalSmall")
	abutton.text:SetSize(20, 10)
	abutton.text:SetJustifyH("LEFT")
	abutton.text:SetPoint("TOPLEFT", abutton.icon, 4, -7)
	
	abutton:SetScript("OnEvent", QuestObjectiveItem_OnEvent)
	abutton:SetScript("OnUpdate", QuestObjectiveItem_OnUpdate)
	abutton:SetScript("OnShow", QuestObjectiveItem_OnShow)
	abutton:SetScript("OnHide", QuestObjectiveItem_OnHide)
	abutton:SetScript("OnEnter", QuestObjectiveItem_OnEnter)
	abutton:SetScript("OnLeave", QuestObjectiveItem_OnLeave)
	abutton:RegisterForClicks("AnyUp")
	abutton:SetAttribute("type","item")
	
	abutton:SetPushedTexture("Interface\\Buttons\\UI-Quickslot-Depress")
	do local tex = abutton:GetPushedTexture()
		tex:SetPoint("TOPLEFT", 0, -1)
		tex:SetPoint("BOTTOMRIGHT", 0, -1)
	end
	abutton:SetHighlightTexture("Interface\\Buttons\\ButtonHilight-Square", "ADD")
	do local tex = abutton:GetHighlightTexture()
		tex:SetPoint("TOPLEFT", 0, -1)
		tex:SetPoint("BOTTOMRIGHT", 0, -1)
	end
	abutton:Hide()
	QGTF.ActiveButton = abutton
	QGT:Masque_AddButton(abutton)
	
	-- Frame locks
	if FRAMELOCK_STATES and FRAMELOCK_STATES.PETBATTLES then
		FRAMELOCK_STATES.PETBATTLES.KalielsTrackerFrame = "hidden"
		FRAMELOCK_STATES.PETBATTLES.KalielsTrackerButtons = "hidden"
	end
end

function QGT:MoveTracker()
	QGTF:ClearAllPoints()
	QGTF:SetPoint(db.anchorPoint, UIParent, db.anchorPoint, db.xOffset, db.yOffset)
	QGTF.directionUp = (db.anchorPoint == "BOTTOMLEFT" or db.anchorPoint == "BOTTOMRIGHT")
	QGTF.anchorLeft = (db.anchorPoint == "LEFT" or db.anchorPoint == "TOPLEFT" or db.anchorPoint == "BOTTOMLEFT")

	local options = QGT.options.args.general.args.sec1.args
	if QGTF.anchorLeft then
		options.xOffset.min = 0
		options.xOffset.max = QGT.screenWidth - trackerWidth
	else
		options.xOffset.min = -(QGT.screenWidth - trackerWidth)
		options.xOffset.max = 0
	end

	if QGTF.directionUp then
		options.yOffset.min = 0
		options.yOffset.max = QGT.screenHeight - 100
	else
		options.yOffset.min = -(QGT.screenHeight - 100)
		options.yOffset.max = 0
	end

	local mod = (db.anchorPoint == "LEFT" or db.anchorPoint == "RIGHT") and abs(db.yOffset) or 0
	options.maxHeight.max = QGT.screenHeight - abs(db.yOffset) - mod
	db.maxHeight = (abs(db.yOffset)+db.maxHeight > QGT.screenHeight-mod) and options.maxHeight.max or db.maxHeight

	QGT:MoveButtons()
end

function QGT:SetBackground()
	local backdrop = {
		bgFile = LSM:Fetch("background", db.bgr),
		edgeFile = LSM:Fetch("border", db.border),
		edgeSize = db.borderThickness,
		insets = { left=db.bgrInset, right=db.bgrInset, top=db.bgrInset, bottom=db.bgrInset }
	}
	local borderColor = db.classBorder and QGT.classColor or db.borderColor
	
	QGTF:SetBackdrop(backdrop)
	QGTF:SetBackdropColor(db.bgrColor.r, db.bgrColor.g, db.bgrColor.b, db.bgrColor.a)
	QGTF:SetBackdropBorderColor(borderColor.r, borderColor.g, borderColor.b, db.borderAlpha)
	
	hbColor = db.hbBorderColor and borderColor or db.hbColor
	OTFButton:GetNormalTexture():SetVertexColor(hbColor.r, hbColor.g, hbColor.b)
	if db.hbOtherButtons then
		OTFHeader.QuestLogButton:GetNormalTexture():SetVertexColor(hbColor.r, hbColor.g, hbColor.b)
		OTFHeader.AchievementsButton:GetNormalTexture():SetVertexColor(hbColor.r, hbColor.g, hbColor.b)
	end
	
	if db.qiBgrBorder then
		QGTF.Buttons:SetBackdrop(backdrop)
		QGTF.Buttons:SetBackdropColor(db.bgrColor.r, db.bgrColor.g, db.bgrColor.b, db.bgrColor.a)
		QGTF.Buttons:SetBackdropBorderColor(borderColor.r, borderColor.g, borderColor.b, db.borderAlpha)
	else
		QGTF.Buttons:SetBackdrop(nil)
	end
	
	if db.bgr == "None" and db.border == "None" then
		QGTF.Scroll:SetHitRectInsets(-150, 0, 0, 1)
	else
		QGTF.Scroll:SetHitRectInsets(-30, 0, 0, 1)
	end
end

function QGT:SetText()
	font = LSM:Fetch("font", db.font)
	fontShadow = db.fontShadow and 1 or 0
	
	-- Title (collapsed)
	OTFHeader.Title:SetFont(font, db.fontSize, db.fontFlag)
	OTFHeader.Title:SetTextColor(0, 0.7, 1)	-- blue
	OTFHeader.Title:SetShadowColor(0, 0, 0, fontShadow)
	--Quests
	QGT:SetHeader(ObjectiveTrackerBlocksFrame.QuestHeader)		
	-- Achievements
	QGT:SetHeader(ObjectiveTrackerBlocksFrame.AchievementHeader)
	-- Scenario
	QGT:SetHeader(ObjectiveTrackerBlocksFrame.ScenarioHeader)
	ScenarioStageBlock.Stage:SetFont(font, db.fontSize+5, db.fontFlag)
	-- Bonus objective
	QGT:SetHeader(BONUS_OBJECTIVE_TRACKER_MODULE.Header)
	
end

function QGT:SetSize()
	local height = 33	-- collapsed
	local mod = 0
	local PTheight = 0
	
	if OTF.BlocksFrame.contentsHeight then
		OTF.BlocksFrame.contentsHeight = round(OTF.BlocksFrame.contentsHeight)
	else
		return
	end
	
	_DBG(" - height = "..OTF.BlocksFrame.contentsHeight)
	if not db.collapsed and not QGT:IsTrackerEmpty() then
		if AUTO_QUEST_POPUP_TRACKER_MODULE.firstBlock then
			mod = mod + (GetNumAutoQuestPopUps() * 68)
		end
		if BONUS_OBJECTIVE_TRACKER_MODULE.firstBlock then
			mod = mod + BONUS_OBJECTIVE_TRACKER_MODULE.blockPadding
		end
		height = OBJECTIVE_TRACKER_HEADER_HEIGHT + OTF.BlocksFrame.contentsHeight + mod
		_DBG(" - "..OBJECTIVE_TRACKER_HEADER_HEIGHT.." + "..OTF.BlocksFrame.contentsHeight.." + "..mod.." = "..height)
		OTF:SetHeight(height)
		OTF.height = height
		
		if height < 2000 and QGTF.Scroll:GetHeight() < 2000 then
			if height > db.maxHeight then
				_DBG("MOVE ... "..QGTF.Scroll.value.." > "..OTF.height.." - "..db.maxHeight)
				if QGTF.Scroll.value > OTF.height-db.maxHeight then
					QGTF.Scroll.value = OTF.height - db.maxHeight
				end
				QGTF.Scroll:SetVerticalScroll(QGTF.Scroll.value)
				height = db.maxHeight
			elseif height <= db.maxHeight then
				if QGTF.Scroll.value > 0 then
					QGTF.Scroll.value = 0
					QGTF.Scroll:SetVerticalScroll(0)
				end
				QGTF.Scroll:UpdateScrollChildRect()
			end
			QGT:MoveButtons()
		end
	else
		OTF.height = height - 10		
		OTF:SetHeight(OTF.height)
		QGTF.Scroll:SetVerticalScroll(0)
	end
	
	QGTF:SetHeight(height)
	QGTF.height = height
end

function QGT:MoveButtons()
	if not InCombatLockdown() then
		local point, xOfs, yOfs
		if QGTF.anchorLeft then
			point = "LEFT"
			xOfs = QGTF:GetRight() and QGTF:GetRight() + db.qiXOffset
		else
			point = "RIGHT"
			xOfs = QGTF:GetLeft() and QGTF:GetLeft() - db.qiXOffset
		end
		local yMod = 2 + (4 - db.bgrInset)
		if QGTF.directionUp and db.maxHeight < QGTF.Buttons:GetHeight() then
			point = "BOTTOM"..point
			yOfs = QGTF:GetBottom() and QGTF:GetBottom() - ((db.qiBgrBorder and (db.bgr ~= "None" or db.border ~= "None")) and 0 or yMod)
		else
			point = "TOP"..point
			yOfs = QGTF:GetTop() and QGTF:GetTop() + ((db.qiBgrBorder and (db.bgr ~= "None" or db.border ~= "None")) and 0 or yMod)
		end
		if xOfs and yOfs then
			QGTF.Buttons:ClearAllPoints()
			QGTF.Buttons:SetPoint(point, UIParent, "BOTTOMLEFT", xOfs, yOfs)
		end
	end
end

--------------------
-- Hook functions --
--------------------

function QGT:SetHooks()
	local bck_ObjectiveTracker_Update = ObjectiveTracker_Update
	ObjectiveTracker_Update = function(reason, id)
		if reason ~= OBJECTIVE_TRACKER_UPDATE_STATIC then
			_DBG("|cffffff00Update ... "..(reason and reason or ""))
		end
		bck_ObjectiveTracker_Update(reason, id)
		if db.collapsed then
			local _, numQuests = GetNumQuestLogEntries()
			local title = ("%d/%d Quests  -  %d Dailies"):format(numQuests, MAX_QUESTS, GetDailyQuestsCompleted())
			OTFHeader.Title:SetText(title)
		end
		if reason == OBJECTIVE_TRACKER_UPDATE_STATIC then
			return
		elseif reason == OBJECTIVE_TRACKER_UPDATE_TASK_ADDED then
			activeTask = id
			
			QGT:ToggleEmptyTracker(true)
		elseif activeTask then
			isInArea, isOnMap, numObjectives = GetTaskInfo(activeTask)
			if not isInArea then
				activeTask = 0
				QGT:ToggleEmptyTracker()
			end
		end
		QGT:SetSize()
	end

	local bck_DEFAULT_OBJECTIVE_TRACKER_MODULE_AddObjective = DEFAULT_OBJECTIVE_TRACKER_MODULE.AddObjective
	function DEFAULT_OBJECTIVE_TRACKER_MODULE:AddObjective(block, objectiveKey, text, lineType, useFullHeight, hideDash, colorStyle)
		local _, _, progress, numHave, numNeed, desc = strfind(text, "^((%d+).*/(%d+)%a?)(.*)")
		if progress and numHave ~= "0" and numNeed ~= numHave then
			text = "|cffc8c800" .. progress .. "|r" .. desc
		end
		local line = bck_DEFAULT_OBJECTIVE_TRACKER_MODULE_AddObjective(QGT, block, objectiveKey, text, lineType, useFullHeight, hideDash, colorStyle)
		if line.Dash then
			line.Dash:SetFont(font, db.fontSize, db.fontFlag)
			line.Dash:SetShadowColor(0, 0, 0, fontShadow)
		end
		if line.Check then
			line.Check:SetSize(db.fontSize-2.5, db.fontSize-2.5)
			line.Check:ClearAllPoints()
			line.Check:SetPoint("LEFT", -db.fontSize*0.2+(db.fontFlag == "None" and 0 or 1), 0)
		end
		return line
	end

	local bck_DEFAULT_OBJECTIVE_TRACKER_MODULE_SetStringText = DEFAULT_OBJECTIVE_TRACKER_MODULE.SetStringText
	function DEFAULT_OBJECTIVE_TRACKER_MODULE:SetStringText(fontString, text, useFullHeight, colorStyle, useHighlight)
		fontString:SetFont(font, db.fontSize, db.fontFlag)
		fontString:SetShadowColor(0, 0, 0, fontShadow)
		fontString:SetWordWrap(false)
		return bck_DEFAULT_OBJECTIVE_TRACKER_MODULE_SetStringText(QGT, fontString, text, useFullHeight, colorStyle, useHighlight)
	end

	hooksecurefunc(DEFAULT_OBJECTIVE_TRACKER_MODULE, "OnBlockHeaderEnter", function(QGT, block)
		if db.blockTooltips and (QGT == QUEST_TRACKER_MODULE or QGT == ACHIEVEMENT_TRACKER_MODULE) then
			GameTooltip:SetOwner(block, "ANCHOR_NONE")
			GameTooltip:ClearAllPoints()
			if QGTF.anchorLeft then
				GameTooltip:SetPoint("TOPLEFT", block, "TOPRIGHT", -2, 0)
			else
				GameTooltip:SetPoint("TOPRIGHT", block, "TOPLEFT", -20, 0)
			end
			if QGT == QUEST_TRACKER_MODULE then
				GameTooltip:SetHyperlink(GetQuestLink(block.questLogIndex))
			else
				GameTooltip:SetHyperlink(GetAchievementLink(block.id))
			end
			GameTooltip:Show()
		end

		if block.fixedTag then
			local colorStyle = OBJECTIVE_TRACKER_COLOR["HeaderHighlight"]
			block.fixedTag:SetBackdropColor(colorStyle.r, colorStyle.g, colorStyle.b)
			block.fixedTag.text:SetTextColor(colorStyle.r, colorStyle.g, colorStyle.b)
		end
	end)

	hooksecurefunc(DEFAULT_OBJECTIVE_TRACKER_MODULE, "OnBlockHeaderLeave", function(QGT, block)
		if block.questCompleted then
			local colorStyle = OBJECTIVE_TRACKER_COLOR["Complete"]
			block.HeaderText:SetTextColor(colorStyle.r, colorStyle.g, colorStyle.b)
			block.HeaderText.colorStyle = colorStyle
		end
		if db.blockTooltips then
			GameTooltip:Hide()
		end

		if block.fixedTag then
			block.fixedTag:SetBackdropColor(1, 1, 1)
			block.fixedTag.text:SetTextColor(1, 1, 1)
		end
	end)

	hooksecurefunc(QUEST_TRACKER_MODULE, "OnBlockHeaderClick", function(QGT, block, mouseButton)
		GameTooltip:Hide()
	end)
	hooksecurefunc(ACHIEVEMENT_TRACKER_MODULE, "OnBlockHeaderClick", function(QGT, block, mouseButton)
		GameTooltip:Hide()
	end)

	local bck_ObjectiveTracker_AddBlock = ObjectiveTracker_AddBlock
	ObjectiveTracker_AddBlock = function(block, forceAdd)
		local blockAdded = bck_ObjectiveTracker_AddBlock(block, forceAdd)
		if blockAdded and (block.module == QUEST_TRACKER_MODULE or block.module == ACHIEVEMENT_TRACKER_MODULE) then
			local button = block.button
			if not button then
				button = CreateFrame("Button", nil, block)
				button:SetAllPoints()
				button:SetScript("OnEnter", function() block.module:OnBlockHeaderEnter(block) end)
				button:SetScript("OnLeave", function() block.module:OnBlockHeaderLeave(block) end)
				button:SetScript("OnClick", function(QGT, btn) block.module:OnBlockHeaderClick(block, btn) end)
				button:RegisterForClicks("AnyUp")
				block.button = button
			end
			button:Show()
		end
		if not db.collapsed and not blockAdded then
			OTF.BlocksFrame:SetHeight(2000)		-- slide
			OTF.BlocksFrame.contentsHeight = 2000
		end
		return blockAdded
	end

	hooksecurefunc(DEFAULT_OBJECTIVE_TRACKER_MODULE, "FreeBlock", function(QGT, block)
		if block.button then
			block.button:Hide()
		end
	end)

	local bck_QUEST_TRACKER_MODULE_SetBlockHeader = QUEST_TRACKER_MODULE.SetBlockHeader
	function QUEST_TRACKER_MODULE:SetBlockHeader(block, text, questLogIndex, isQuestComplete)
		local _, level, _, _, _, _, frequency, questID = GetQuestLogTitle(questLogIndex)
		local tagID, tagName = GetQuestTagInfo(questID)
		text = QGT:CreateQuestTag(level, tagID, frequency) .. text
		bck_QUEST_TRACKER_MODULE_SetBlockHeader(QGT, block, text, questLogIndex, isQuestComplete)
		if block.questCompleted then
			local colorStyle = OBJECTIVE_TRACKER_COLOR["Complete"]
			block.HeaderText:SetTextColor(colorStyle.r, colorStyle.g, colorStyle.b)
			block.HeaderText.colorStyle = colorStyle
		end

		local link, item, charges, showItemWhenComplete = GetQuestLogSpecialItemInfo(questLogIndex)
		if item and (not isQuestComplete or showItemWhenComplete) then
			QGTF.Buttons.num = QGTF.Buttons.num + 1
			CreateFixedTag(block)
			local button = CreateFixedButton(block)
			if InCombatLockdown() then
				_DBG(" - STOP Create button")
				stopUpdate = true
			else
				button.questLogIndex = questLogIndex
				button.charges = charges
				button.rangeTimer = -1
				button.item = item
				button.link = link
				SetItemButtonTexture(button, item)
				SetItemButtonCount(button, charges)
				QuestObjectiveItem_UpdateCooldown(button)
				button:SetAttribute("item", link)
			end
		else
			RemoveFixedButton(block)
		end
	end

	function CreateFixedTag(block)
		block.itemButton:Hide()

		local tag = block.fixedTag
		if not tag then
			local numFreeButtons = #freeTags
			if numFreeButtons > 0 then
				tag = freeTags[numFreeButtons]
				tremove(freeTags, numFreeButtons)
				tag:SetParent(block)
			else
				tag = CreateFrame("Frame", nil, block)
				tag:SetSize(26, 26)
				tag:SetBackdrop({ bgFile = "Interface\\AddOns\\QuestGuru\\Graphics\\bgr_quest-item" })
				tag.text = tag:CreateFontString(nil, "ARTWORK", "NumberFontNormal")
				tag.text:SetPoint("CENTER", 1, -1)
			end
			tag:SetPoint("TOPRIGHT")
			tag:Show()
			block.fixedTag = tag
		end

		tag:SetBackdropColor(1, 1, 1)
		tag.text:SetTextColor(1, 1, 1)
		tag.text:SetText(QGTF.Buttons.num)
	end

	function CreateFixedButton(block)
		local questID = block.id
		local button = GetFixedButton(questID)
		if not button then
			if InCombatLockdown() then
				return nil
			end

			local numFreeButtons = #freeButtons
			if numFreeButtons > 0 then
				_DBG(" - USE button "..questID)
				button = freeButtons[numFreeButtons]
				tremove(freeButtons, numFreeButtons)
			else
				_DBG(" - CREATE button "..questID)
				local name = addonName.."FixedButton"..questID
				button = CreateFrame("Button", name, QGTF.Buttons, "SecureActionButtonTemplate")
				button:SetSize(26, 26)

				button.icon = button:CreateTexture(name.."Icon", "BORDER")
				button.icon:SetAllPoints()

				button.Count = button:CreateFontString(name.."Count", "BORDER", "NumberFontNormal")
				button.Count:SetJustifyH("RIGHT")
				button.Count:SetPoint("BOTTOMRIGHT", button.icon, -3, 2)

				button.Cooldown = CreateFrame("Cooldown", name.."Cooldown", button, "CooldownFrameTemplate")
				button.Cooldown:SetAllPoints()

				button.HotKey = button:CreateFontString(name.."HotKey", "ARTWORK", "NumberFontNormalSmallGray")
				button.HotKey:SetSize(29, 10)
				button.HotKey:SetJustifyH("RIGHT")
				button.HotKey:SetText(RANGE_INDICATOR)
				button.HotKey:SetPoint("TOPRIGHT", button.icon, 2, -2)

				button.text = button:CreateFontString(name.."Text", "ARTWORK", "NumberFontNormalSmall")
				button.text:SetSize(29, 10)
				button.text:SetJustifyH("LEFT")
				button.text:SetPoint("TOPLEFT", button.icon, 1, -3)

				button:SetScript("OnEvent", QuestObjectiveItem_OnEvent)
				button:SetScript("OnUpdate", QuestObjectiveItem_OnUpdate)
				button:SetScript("OnShow", QuestObjectiveItem_OnShow)
				button:SetScript("OnHide", QuestObjectiveItem_OnHide)
				button:SetScript("OnEnter", QuestObjectiveItem_OnEnter)
				button:SetScript("OnLeave", QuestObjectiveItem_OnLeave)

				button:RegisterForClicks("AnyUp")
				button:SetAttribute("type","item")

				button:SetNormalTexture("Interface\\Buttons\\UI-Quickslot2")
				do local tex = button:GetNormalTexture()
					tex:ClearAllPoints()
					tex:SetPoint("CENTER")
					tex:SetSize(44, 44)
				end
				button:SetPushedTexture("Interface\\Buttons\\UI-Quickslot-Depress")
				button:SetHighlightTexture("Interface\\Buttons\\ButtonHilight-Square", "ADD")
				button:SetFrameLevel(QGTF:GetFrameLevel() + 1)
				button:Hide()	-- Cooldown init

				QGT:Masque_AddButton(button)
			end
			button:Show()
			fixedButtons[questID] = button
			QGTF.Buttons.reanchor = true
		end
		button.block = block
		button.num = QGTF.Buttons.num
		button.text:SetText(QGTF.Buttons.num)
		button:SetAlpha(1)
		if QGTF.ActiveButton.questID == questID then
			QGT:UpdateActiveButton(questID)
		end
		return button
	end

	function RemoveFixedButton(block)
		if block then
			local tag = block.fixedTag
			if tag then
				tinsert(freeTags, tag)
				block.fixedTag = nil
				tag:Hide()
			end
			local questID = block.id
			local button = GetFixedButton(questID)
			if button then
				button:SetAlpha(0)
				if InCombatLockdown() then
					_DBG(" - STOP Remove button")
					stopUpdate = true
				else
					_DBG(" - REMOVE button "..questID)
					tinsert(freeButtons, button)
					fixedButtons[questID] = nil
					button:Hide()
					QGTF.Buttons.reanchor = true
				end
				QGTF.ActiveButton.text:SetText("")
			end
			if QGTF.Buttons.timerID == questID then
				QGTF.Buttons.timerID = nil
			end
		else
			for questID, button in pairs(fixedButtons) do
				_DBG(" - REMOVE button "..questID)
				tinsert(freeButtons, button)
				fixedButtons[questID] = nil
				button:Hide()
			end
		end
	end

	function GetFixedButton(questID)
		if fixedButtons[questID] then
			return fixedButtons[questID]
		else
			return nil
		end
	end

	function QuestObjectiveItem_OnEnter(QGT)
		if QGTF.anchorLeft then
			GameTooltip:SetOwner(QGT, "ANCHOR_RIGHT", 2)
		else
			GameTooltip:SetOwner(QGT, "ANCHOR_LEFT", -2)
		end
		GameTooltip:SetQuestLogSpecialItem(QGT.questLogIndex)

		local block = QGT.block
		local colorStyle = OBJECTIVE_TRACKER_COLOR["HeaderHighlight"]
		block.HeaderText:SetTextColor(colorStyle.r, colorStyle.g, colorStyle.b)
		block.HeaderText.colorStyle = colorStyle

		if block.fixedTag then
			block.fixedTag:SetBackdropColor(colorStyle.r, colorStyle.g, colorStyle.b)
			block.fixedTag.text:SetTextColor(colorStyle.r, colorStyle.g, colorStyle.b)
		end
	end

	function QuestObjectiveItem_OnLeave(QGT)
		GameTooltip:Hide()

		local colorStyle
		local block = QGT.block
		if block.questCompleted then
			colorStyle = OBJECTIVE_TRACKER_COLOR["Complete"]
		else
			colorStyle = OBJECTIVE_TRACKER_COLOR["Header"]
		end
		block.HeaderText:SetTextColor(colorStyle.r, colorStyle.g, colorStyle.b)
		block.HeaderText.colorStyle = colorStyle

		if block.fixedTag then
			block.fixedTag:SetBackdropColor(1, 1, 1)
			block.fixedTag.text:SetTextColor(1, 1, 1)
		end
	end

	function QuestObjectiveItem_OnUpdate(QGT, elapsed)
		local rangeTimer = QGT.rangeTimer
		if rangeTimer then
			rangeTimer = rangeTimer - elapsed
			if rangeTimer <= 0 then
				local link, item, charges, showItemWhenComplete = GetQuestLogSpecialItemInfo(QGT.questLogIndex)
				if not charges or charges ~= QGT.charges then
					ObjectiveTracker_Update(OBJECTIVE_TRACKER_UPDATE_MODULE_QUEST)
					return
				end
				local hotkey = QGT.HotKey
				local valid = IsQuestLogSpecialItemInRange(QGT.questLogIndex)
				if hotkey:GetText() == RANGE_INDICATOR then
					if valid == 0 then
						hotkey:Show()
						hotkey:SetVertexColor(1.0, 0.1, 0.1)
					elseif valid == 1 then
						hotkey:Show()
						hotkey:SetVertexColor(0.6, 0.6, 0.6)
					else
						hotkey:Hide()
					end
				else
					if valid == 0 then
						hotkey:SetVertexColor(1.0, 0.1, 0.1)
					else
						hotkey:SetVertexColor(0.6, 0.6, 0.6)
					end
				end
				rangeTimer = TOOLTIP_UPDATE_TIME
			end
			QGT.rangeTimer = rangeTimer
		end

		if db.qiActiveButton and not InCombatLockdown() then
			if QGTF.Buttons.timerID then
				if QGTF.Buttons.timerID ~= QGT.block.id then
					return
				end
			else
				QGTF.Buttons.timerID = QGT.block.id
			end
			if QGTF.Buttons.timer > 50 then
				QGT:UpdateActiveButton()
			else
				--_DBG("... "..QGTF.Buttons.timer.." ... "..QGT.block.id)
				QGTF.Buttons.timer = QGTF.Buttons.timer + TOOLTIP_UPDATE_TIME
			end
		end
	end

	local bck_QUEST_TRACKER_MODULE_OnFreeBlock = QUEST_TRACKER_MODULE.OnFreeBlock
	function QUEST_TRACKER_MODULE:OnFreeBlock(block)
		RemoveFixedButton(block)
		bck_QUEST_TRACKER_MODULE_OnFreeBlock(QGT, block)
	end

	local bck_QUEST_TRACKER_MODULE_Update = QUEST_TRACKER_MODULE.Update
	function QUEST_TRACKER_MODULE:Update()
		QGTF.Buttons.num = 0
		bck_QUEST_TRACKER_MODULE_Update()
	end

	hooksecurefunc(QUEST_TRACKER_MODULE, "EndLayout", function(QGT, isStaticReanchor)
		if not InCombatLockdown() and QGTF.Buttons.reanchor then
			_DBG(" - REANCHOR buttons")
			local questID, block, yOfs
			local numButtons = 0
			local contentsHeight = 0
			for i=1, GetNumQuestWatches() do
				questID, _ = GetQuestWatchInfo(i)
				block = QUEST_TRACKER_MODULE.usedBlocks[questID]
				if block and block.itemButton then
					numButtons = numButtons + 1
					yOfs = -(contentsHeight + 7)
					contentsHeight = contentsHeight + 26 + 3
					fixedButtons[questID]:SetPoint("TOP", 0, yOfs)
				end
			end
			QGTF.Buttons:SetHeight(contentsHeight + 7 + 4)
			QGTF.Buttons.height = contentsHeight
			QGTF.Buttons.reanchor = false

			QGT:UpdateActiveButton()
		end
		if db.collapsed or QGTF.Buttons.height == 0 then
			QGTF.Buttons:SetAlpha(0)
		else
			QGTF.Buttons:SetAlpha(1)
		end
	end)

	local bck_BONUS_OBJECTIVE_TRACKER_MODULE_AddProgressBar = BONUS_OBJECTIVE_TRACKER_MODULE.AddProgressBar
	function BONUS_OBJECTIVE_TRACKER_MODULE:AddProgressBar(block, line, questID, finished)
		local progressBar = bck_BONUS_OBJECTIVE_TRACKER_MODULE_AddProgressBar(QGT, block, line, questID, finished)
		progressBar.Bar.Label:SetPoint("CENTER")
		progressBar.Bar.Label:SetFont(font, 13, db.fontFlag)
		progressBar.Bar.Label:SetShadowColor(0, 0, 0, fontShadow)
		progressBar.Bar.Label:Show()
		progressBar.Bar:SetScript("OnEnter", function() BonusObjectiveTracker_OnBlockEnter(block) end)
		progressBar.Bar:SetScript("OnLeave", function() BonusObjectiveTracker_OnBlockLeave(block) end)
		return progressBar
	end

	hooksecurefunc("BonusObjectiveTracker_ShowRewardsTooltip", function(block)
		if QGTF.anchorLeft then
			GameTooltip:ClearAllPoints()
			GameTooltip:SetPoint("TOPLEFT", block, "TOPRIGHT", 10, 0)
		end
	end)

	function ScenarioObjectiveStageBlock_OnEnter2(QGT)
		ScenarioObjectiveStageBlock_OnEnter(QGT)
		if QGTF.anchorLeft then
			GameTooltip:ClearAllPoints()
			GameTooltip:SetPoint("LEFT", QGT, "RIGHT", 42, 0)
		end
	end
	ScenarioStageBlock:SetScript("OnEnter", ScenarioObjectiveStageBlock_OnEnter2)

	function ObjectiveTracker_Collapse()
		_DBG("--------------------------------")
		_DBG("COLLAPSE")
		OTF.collapsed = true
		QGT.db.profile.collapsed = OTF.collapsed
		OTF.BlocksFrame:Hide()
		OTFButton:GetNormalTexture():SetTexCoord(0, 0.5, 0, 0.25)
		OTFHeader.Title:Show()
		CloseDropDownMenus()
		animTask = false		
	end

	function ObjectiveTracker_Expand()
		_DBG("--------------------------------")
		_DBG("EXPAND")
		OTF.collapsed = nil
		QGT.db.profile.collapsed = OTF.collapsed
		OTF.BlocksFrame:Show()
		OTFButton:GetNormalTexture():SetTexCoord(0, 0.5, 0.25, 0.5)
		OTFHeader.Title:Hide()
	end

	hooksecurefunc("BonusObjectiveTracker_OnBlockAnimOutFinished", function(QGT)
		animTask = false
		
	end)

	hooksecurefunc("BonusObjectiveTracker_SetBlockState", function(block, state, force)
		if state == "PRESENT" then
			_DBG(" - "..state)
			activeTask = block.id
		elseif state == "LEAVING" and activeTask > 0 then
			_DBG(" - "..state)
			animTask = true
			activeTask = 0
			if QGT:IsTrackerEmpty(true) then
				OTF.BlocksFrame.contentsHeight = OTF.BlocksFrame.contentsHeight + paddingBottom
			end
		end		
	end)

	hooksecurefunc("ExtraActionBar_Update", function()
		if QGTF.ActiveButton.questID then
			local bar = ExtraActionBarFrame
			bar.outro:Stop()
			bar:SetAlpha(1)
		end
	end)

	function QuestActiveButton_UpdateHotkey()
		local key = GetBindingKey("EXTRAACTIONBUTTON1")
		if db.keyBindMinimize == key then
			SetOverrideBinding(QGTF, false, db.keyBindMinimize, nil)
			db.keyBindMinimize = ""
		end
		if db.qiActiveButton then
			local abutton = QGTF.ActiveButton
			local hotkey = abutton.HotKey
			local text = GetBindingText(key, 1)
			ClearOverrideBindings(abutton)
			if key then
				SetOverrideBindingClick(abutton, false, key, abutton:GetName())
			end
			if text == "" then
				hotkey:SetText(RANGE_INDICATOR)
				hotkey:Hide()
			else
				hotkey:SetText(text)
				hotkey:Show()
			end
		end
	end
end

function QGT:UpdateActiveButton(id)
	if not db.qiActiveButton then
		return
	end

	local button
	local abutton = QGTF.ActiveButton
	local autoShowTooltip = false

	if id then
		button = GetFixedButton(id)
		if GameTooltip:IsShown() and GameTooltip:GetOwner() == abutton then
			QuestObjectiveItem_OnLeave(abutton)
			autoShowTooltip = true
		end
		abutton.block = button.block
		abutton.text:SetText(button.num)
		if autoShowTooltip then
			QuestObjectiveItem_OnEnter(abutton)
		end
		return
	end

	if InCombatLockdown() then
		return
	end

	local closestQuestID
	local minDistSqr = 22500
	local bar = ExtraActionBarFrame

	for questID, button in pairs(fixedButtons) do
		if QuestHasPOIInfo(questID) then
			local distSqr, onContinent = GetDistanceSqToQuest(button.questLogIndex)
			if onContinent and distSqr <= minDistSqr then
				minDistSqr = distSqr
				closestQuestID = questID
			end
		end
	end

	if closestQuestID then
		button = GetFixedButton(closestQuestID)
		if abutton.questID ~= closestQuestID or not bar:IsShown() then
			if GameTooltip:IsShown() and GameTooltip:GetOwner() == abutton then
				QuestObjectiveItem_OnLeave(abutton)
				autoShowTooltip = true
			end

			bar:Show()
			bar.button:SetAlpha(0)

			abutton:Show()
			abutton.block = button.block
			abutton.questID = closestQuestID
			abutton.questLogIndex = button.questLogIndex
			abutton.charges = button.charges
			abutton.rangeTimer = button.rangeTimer
			SetItemButtonTexture(abutton, button.item)
			SetItemButtonCount(abutton, button.charges)
			QuestObjectiveItem_UpdateCooldown(abutton)
			abutton:SetAttribute("item", button.link)
			QuestActiveButton_UpdateHotkey()

			UIParent_ManageFramePositions()
			bar.outro:Stop()
			bar:SetAlpha(1)

			if autoShowTooltip then
				QuestObjectiveItem_OnEnter(abutton)
			end
		end
		abutton.text:SetText(button.num)
	elseif bar:IsShown() then
		bar.button:SetAlpha(1)
		if not HasExtraActionBar() then
			bar.intro:Stop()
			bar:SetAlpha(0)
			bar:Hide()
		end

		abutton:Hide()
		ClearOverrideBindings(abutton)
	end
	QGTF.Buttons.timer = 0
end

function QGT:ToggleTrackAllQuests()
	local firstQuestID
	local numWatches = GetNumQuestWatches()
	local numEntries, numQuests = GetNumQuestLogEntries()
	if numWatches > 0 then
		for questLogIndex=1, numEntries do
			local _, _, _, isHeader = GetQuestLogTitle(questLogIndex)
			if not isHeader then
				RemoveQuestWatch(questLogIndex)
			end
		end
	else
		for questLogIndex=1, numEntries do
			local _, _, _, isHeader, _, _, _, questID = GetQuestLogTitle(questLogIndex)
			if not isHeader and not IsQuestWatched(questLogIndex) then
				AddQuestWatch(questLogIndex, true)
				firstQuestID = questID
			end
		end
		QuestSuperTracking_OnQuestTracked(firstQuestID)
		ObjectiveTracker_Update()	-- preventive
	end
end

function QGT:CreateQuestTag(level, questTag, frequency)
	local tag = ""
	
	if level == -1 then
		level = "*"
	else
		level = tostring(level)
	end
	
	if questTag == ELITE then
		tag = "+"
	elseif questTag == QUEST_TAG_GROUP then
		tag = "g"
	elseif questTag == QUEST_TAG_PVP then
		tag = "pvp"
	elseif questTag == QUEST_TAG_DUNGEON then
		tag = "d"
	elseif questTag == QUEST_TAG_HEROIC then
		tag = "hc"
	elseif questTag == QUEST_TAG_RAID then
		tag = "r"
	elseif questTag == QUEST_TAG_RAID10 then
		tag = "r10"
	elseif questTag == QUEST_TAG_RAID25 then
		tag = "r25"
	elseif questTag == QUEST_TAG_SCENARIO then
		tag = "s"
	elseif questTag == QUEST_TAG_ACCOUNT then
		tag = "a"
	elseif questTag == QUEST_TAG_LEGENDARY then
		tag = "leg"
	end
	
	if frequency == 2 then
		tag = tag.."!"		-- daily quest
	elseif frequency == 3 then
		tag = tag.."!!"	-- weekly quest
	end
	
	if tag ~= "" then
		tag = ("|cff00b3ff%s|r"):format(tag)
	end
	
	tag = ("[%s%s] "):format(level, tag)
	return tag
end

function QGT:SetHeader(obj)
	obj.Text:SetFont(font, db.fontSize+2, db.fontFlag)
	obj.Text:SetShadowColor(0, 0, 0, fontShadow)
	obj.Text:SetPoint("LEFT", 4, 0)
	obj.animateReason = 0
end

function QGT:ToggleEmptyTracker(added)
	local alpha, mouse = 1, true
	if QGT:IsTrackerEmpty() then
		if not db.collapsed then
			ObjectiveTracker_MinimizeButton_OnClick()										
		end
		OTFButton:GetNormalTexture():SetTexCoord(0, 0.5, 0.5, 0.75)
		if db.hideEmptyTracker then
			alpha = 0
			mouse = false
		end
	else
		if db.collapsed then
			if added then
				ObjectiveTracker_MinimizeButton_OnClick()
			else
				OTFButton:GetNormalTexture():SetTexCoord(0, 0.5, 0, 0.25)
			end
		else
			OTFButton:GetNormalTexture():SetTexCoord(0, 0.5, 0.25, 0.5)
		end
	end
	QGTF:SetAlpha(alpha)
	OTFHeader.MinimizeButton:EnableMouse(mouse)
	if db.hbOtherButtons then
		OTFHeader.QuestLogButton:EnableMouse(mouse)
		OTFHeader.AchievementsButton:EnableMouse(mouse)
	end
end

function QGT:IsTrackerEmpty(noaddon)
	local result = (GetNumQuestWatches() == 0 and
			GetNumAutoQuestPopUps() == 0 and
			GetNumTrackedAchievements() == 0 and
			activeTask == 0 and
			not C_Scenario.IsInScenario())
	
	return result
end


QGT:OnInitialize()